try:
    from urllib.parse import urlencode
except ImportError:
    # Python 2
    from urllib import urlencode

from elasticapm.utils import get_url_dict
from elasticapm.utils.wsgi import get_environ, get_headers


def get_data_from_request(request):
    body = None
    if request.data:
        body = request.data
    elif request.form:
        body = urlencode(request.form)

    result = {
        'body': body,
        'env': dict(get_environ(request.environ)),
        'headers': dict(
            get_headers(request.environ),
        ),
        'method': request.method,
        'socket': {
            'remote_address': request.environ.get('REMOTE_ADDR'),
            'encrypted': request.is_secure
        },
        'cookies': request.cookies,
    }

    result['url'] = get_url_dict(request.url)

    return result



import sqlite3

import mock

import elasticapm.instrumentation.control
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class InstrumentSQLiteTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        elasticapm.instrumentation.control.instrument()

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_connect(self, should_collect):
        should_collect.return_value = False
        self.client.begin_transaction("transaction.test")

        conn = sqlite3.connect(":memory:")
        cursor = conn.cursor()

        cursor.execute("""CREATE TABLE testdb (id integer, username text)""")
        cursor.execute("""INSERT INTO testdb VALUES (1, "Ron")""")
        cursor.execute("""DROP TABLE testdb""")

        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', 'sqlite3.connect :memory:',
                               'CREATE TABLE', 'INSERT INTO testdb',
                               'DROP TABLE']

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(traces[0]['name'], 'sqlite3.connect :memory:')
        self.assertEqual(traces[0]['type'], 'db.sqlite.connect')

        self.assertEqual(traces[1]['name'], 'CREATE TABLE')
        self.assertEqual(traces[1]['type'], 'db.sqlite.sql')

        self.assertEqual(traces[2]['name'], 'INSERT INTO testdb')
        self.assertEqual(traces[2]['type'], 'db.sqlite.sql')

        self.assertEqual(traces[3]['name'], 'DROP TABLE')
        self.assertEqual(traces[3]['type'], 'db.sqlite.sql')

        self.assertEqual(traces[4]['name'], 'transaction')
        self.assertEqual(traces[4]['type'], 'transaction')

        self.assertEqual(len(traces), 5)



import logging

from elasticapm.handlers.logging import LoggingHandler
from elasticapm.utils.stacks import iter_stack_frames
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class LoggingIntegrationTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient(include_paths=['tests', 'elasticapm'])
        self.handler = LoggingHandler(self.client)
        self.logger = logging.getLogger(__name__)
        self.logger.handlers = []
        self.logger.addHandler(self.handler)

    def test_logger_basic(self):
        self.logger.error('This is a test error')

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['log']['logger_name'], __name__)
        self.assertEquals(event['log']['level'], "error")
        self.assertEquals(event['log']['message'], 'This is a test error')
        self.assertFalse('stacktrace' in event['log'])
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test error')

    def test_logger_warning(self):
        self.logger.warning('This is a test warning')
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['log']['logger_name'], __name__)
        self.assertEquals(event['log']['level'], "warning")
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test warning')

    def test_logger_extra_data(self):
        self.logger.info('This is a test info with a url', extra=dict(
            data=dict(
                url='http://example.com',
            ),
        ))
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['context']['custom']['url'], 'http://example.com')
        self.assertFalse('stacktrace' in event['log'])
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test info with a url')

    def test_logger_exc_info(self):
        try:
            raise ValueError('This is a test ValueError')
        except ValueError:
            self.logger.info('This is a test info with an exception', exc_info=True)

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]

        # self.assertEquals(event['message'], 'This is a test info with an exception')
        self.assertTrue('exception' in event)
        self.assertTrue('stacktrace' in event['exception'])
        exc = event['exception']
        self.assertEquals(exc['type'], 'ValueError')
        self.assertEquals(exc['message'], 'ValueError: This is a test ValueError')
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['message'], 'This is a test info with an exception')

    def test_message_params(self):
        self.logger.info('This is a test of %s', 'args')
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['message'], 'This is a test of args')
        self.assertEquals(event['log']['param_message'], 'This is a test of %s')

    def test_record_stack(self):
        self.logger.info('This is a test of stacks', extra={'stack': True})
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        frames = event['log']['stacktrace']
        self.assertNotEquals(len(frames), 1)
        frame = frames[0]
        self.assertEquals(frame['module'], __name__)
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test of stacks')
        self.assertEquals(event['culprit'], 'tests.handlers.logging.logging_tests.test_record_stack')
        self.assertEquals(event['log']['message'], 'This is a test of stacks')

    def test_no_record_stack(self):
        self.logger.info('This is a test of no stacks', extra={'stack': False})
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event.get('culprit'), None)
        self.assertEquals(event['log']['message'], 'This is a test of no stacks')
        self.assertFalse('stacktrace' in event['log'])
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test of no stacks')

    def test_explicit_stack(self):
        self.logger.info('This is a test of stacks', extra={'stack': iter_stack_frames()})
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertTrue('culprit' in event, event)
        self.assertEquals(event['culprit'], 'tests.handlers.logging.logging_tests.test_explicit_stack')
        self.assertTrue('message' in event['log'], event)
        self.assertEquals(event['log']['message'], 'This is a test of stacks')
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test of stacks')
        self.assertTrue('stacktrace' in event['log'])

    def test_extra_culprit(self):
        self.logger.info('This is a test of stacks', extra={'culprit': 'foo.bar'})
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['culprit'], 'foo.bar')

    def test_logger_exception(self):
        try:
            raise ValueError('This is a test ValueError')
        except ValueError:
            self.logger.exception('This is a test with an exception', extra={'stack': True})

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]

        self.assertEquals(event['log']['message'], 'This is a test with an exception')
        self.assertTrue('stacktrace' in event['log'])
        self.assertTrue('exception' in event)
        exc = event['exception']
        self.assertEquals(exc['type'], 'ValueError')
        self.assertEquals(exc['message'], 'ValueError: This is a test ValueError')
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['message'], 'This is a test with an exception')


class LoggingHandlerTest(TestCase):
    def test_client_arg(self):
        client = get_tempstoreclient(include_paths=['tests'])
        handler = LoggingHandler(client)
        self.assertEquals(handler.client, client)

    def test_client_kwarg(self):
        client = get_tempstoreclient(include_paths=['tests'])
        handler = LoggingHandler(client=client)
        self.assertEquals(handler.client, client)

    def test_invalid_first_arg_type(self):
        self.assertRaises(ValueError, LoggingHandler, object)



import mock
import pytest
import urllib3.poolmanager
from urllib3.exceptions import MaxRetryError, TimeoutError
from urllib3_mock import Responses

from elasticapm.transport.base import TransportException
from elasticapm.transport.http_urllib3 import (AsyncUrllib3Transport,
                                               Urllib3Transport)

try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse



responses = Responses('urllib3')

@responses.activate
def test_send():
    transport = Urllib3Transport(urlparse.urlparse('http://localhost'))
    responses.add('POST', '/', status=202,
                  adding_headers={'Location': 'http://example.com/foo'})
    url = transport.send('x', {})
    assert url == 'http://example.com/foo'


@responses.activate
def test_timeout():
    transport = Urllib3Transport(urlparse.urlparse('http://localhost'))
    responses.add('POST', '/', status=202,
                  body=MaxRetryError(None, None, reason=TimeoutError()))
    with pytest.raises(TransportException) as exc_info:
        transport.send('x', {})
    assert 'timeout' in str(exc_info.value)


@responses.activate
def test_http_error():
    url, status, body = (
        'http://localhost:9999', 418, 'Nothing'
    )
    transport = Urllib3Transport(urlparse.urlparse(url))
    responses.add('POST', '/', status=status, body=body)

    with pytest.raises(TransportException) as exc_info:
        transport.send('x', {})
    for val in (status, body):
        assert str(val) in str(exc_info.value)


@responses.activate
def test_generic_error():
    url, status, message, body = (
        'http://localhost:9999', 418, "I'm a teapot", 'Nothing'
    )
    transport = Urllib3Transport(urlparse.urlparse(url))
    responses.add('POST', '/', status=status, body=Exception('Oopsie'))
    with pytest.raises(TransportException) as exc_info:
        transport.send('x', {})
    assert 'Oopsie' in str(exc_info.value)


def test_http_proxy_environment_variable():
    with mock.patch.dict('os.environ', {'HTTP_PROXY': 'http://example.com'}):
        transport = Urllib3Transport(urlparse.urlparse('http://localhost:9999'))
        assert isinstance(transport.http, urllib3.ProxyManager)


def test_https_proxy_environment_variable():
    with mock.patch.dict('os.environ', {'HTTPS_PROXY': 'https://example.com'}):
        transport = Urllib3Transport(urlparse.urlparse('http://localhost:9999'))
        assert isinstance(transport.http, urllib3.poolmanager.ProxyManager)


def test_https_proxy_environment_variable_is_preferred():
    with mock.patch.dict('os.environ', {'HTTPS_PROXY': 'https://example.com',
                                        'HTTP_PROXY': 'http://example.com'}):
        transport = Urllib3Transport(urlparse.urlparse('http://localhost:9999'))
        assert isinstance(transport.http, urllib3.poolmanager.ProxyManager)
        assert transport.http.proxy.scheme == 'https'



"""
elasticapm
~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

__all__ = ('VERSION', 'Client')

try:
    VERSION = __import__('pkg_resources') \
        .get_distribution('elastic-apm').version
except Exception as e:
    VERSION = 'unknown'

from elasticapm.base import *
from elasticapm.conf import *
from elasticapm.traces import *



# -*- coding: utf-8 -*-

from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule


class _TestInstrumentNonExistingFunctionOnModule(AbstractInstrumentedModule):
    name = "test_non_existing_function_instrumentation"
    instrument_list = [
        ("os.path", "non_existing_function")
    ]


class _TestInstrumentNonExistingMethod(AbstractInstrumentedModule):
    name = "test_non_existing_method_instrumentation"
    instrument_list = [
        ("dict", "non_existing_method")
    ]


def test_instrument_nonexisting_method_on_module():
    _TestInstrumentNonExistingFunctionOnModule().instrument()


def test_instrument_nonexisting_method():
    _TestInstrumentNonExistingMethod().instrument()



# -*- coding: utf-8 -*-
from elasticapm.instrumentation.packages.mysql import extract_signature
from tests.utils.compat import TestCase


class ExtractSignatureTest(TestCase):
    def test_insert(self):
        sql = """INSERT INTO `mytable` (id, name) VALUE ('2323', 'Ron')"""
        actual = extract_signature(sql)

        self.assertEqual("INSERT INTO mytable", actual)

    def test_update(self):
        sql = """UPDATE `mytable` set name='Ron' WHERE id = 2323"""
        actual = extract_signature(sql)

        self.assertEqual("UPDATE mytable", actual)

    def test_delete(self):
        sql = """DELETE FROM `mytable` WHERE id = 2323"""
        actual = extract_signature(sql)

        self.assertEqual("DELETE FROM mytable", actual)

    def test_select_simple(self):
        sql = """SELECT `id`, `name` FROM `mytable` WHERE id = 2323"""
        actual = extract_signature(sql)

        self.assertEqual("SELECT FROM mytable", actual)

    def test_select_with_entity_quotes(self):
        sql = """SELECT `id`, `name` FROM `mytable` WHERE id = 2323"""
        actual = extract_signature(sql)

        self.assertEqual("SELECT FROM mytable", actual)

    def test_select_with_difficult_values(self):
        sql = """SELECT id, 'some \\'name' + " from Denmark" FROM `mytable` WHERE id = 2323"""
        actual = extract_signature(sql)

        self.assertEqual("SELECT FROM mytable", actual)

    def test_select_with_difficult_table_name(self):
        sql = "SELECT id FROM `myta\n-æøåble` WHERE id = 2323"""
        actual = extract_signature(sql)

        self.assertEqual("SELECT FROM myta\n-æøåble", actual)

    def test_select_subselect(self):
        sql = """SELECT id, name FROM (
                SELECT id, "not a FROM ''value" FROM mytable WHERE id = 2323
        ) LIMIT 20"""
        actual = extract_signature(sql)

        self.assertEqual("SELECT FROM mytable", actual)

    def test_select_subselect_with_alias(self):
        sql = """
        SELECT count(*)
        FROM (
            SELECT count(id) AS some_alias, some_column
            FROM mytable
            GROUP BY some_colun
            HAVING count(id) > 1
        ) AS foo
        """
        actual = extract_signature(sql)

        self.assertEqual("SELECT FROM mytable", actual)

    def test_select_with_multiple_tables(self):
        sql = """SELECT count(table2.id)
            FROM table1, table2, table2
            WHERE table2.id = table1.table2_id
        """
        actual = extract_signature(sql)
        self.assertEqual("SELECT FROM table1", actual)

    def test_select_with_invalid_literal(self):
        sql = "SELECT \"neverending literal FROM (SELECT * FROM ..."""
        actual = extract_signature(sql)

        self.assertEqual("SELECT FROM", actual)

    def test_savepoint(self):
        sql = """SAVEPOINT x_asd1234"""
        actual = extract_signature(sql)

        self.assertEqual("SAVEPOINT", actual)

    def test_begin(self):
        sql = """BEGIN"""
        actual = extract_signature(sql)

        self.assertEqual("BEGIN", actual)

    def test_create_index_with_name(self):
        sql = """CREATE INDEX myindex ON mytable"""
        actual = extract_signature(sql)

        self.assertEqual("CREATE INDEX", actual)

    def test_create_index_without_name(self):
        sql = """CREATE INDEX ON mytable"""
        actual = extract_signature(sql)

        self.assertEqual("CREATE INDEX", actual)

    def test_drop_table(self):
        sql = """DROP TABLE mytable"""
        actual = extract_signature(sql)

        self.assertEqual("DROP TABLE", actual)

    def test_multi_statement_sql(self):
        sql = """CREATE TABLE mytable; SELECT * FROM mytable; DROP TABLE mytable"""
        actual = extract_signature(sql)

        self.assertEqual("CREATE TABLE", actual)



"""
elasticapm.contrib.pylons
~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
from elasticapm.base import Client
from elasticapm.middleware import ElasticAPM as Middleware


def list_from_setting(config, setting):
    value = config.get(setting)
    if not value:
        return None
    return value.split()


class ElasticAPM(Middleware):
    def __init__(self, app, config, client_cls=Client):
        client = client_cls(
            servers=list_from_setting(config, 'elasticapm.servers'),
            timeout=config.get('elasticapm.timeout'),
            name=config.get('elasticapm.name'),
            app_name=config.get('elasticapm.app_name'),
            secret_token=config.get('elasticapm.secret_token'),
            include_paths=list_from_setting(config, 'elasticapm.include_paths'),
            exclude_paths=list_from_setting(config, 'elasticapm.exclude_paths'),
        )
        super(ElasticAPM, self).__init__(app, client)



"""
elasticapm.core.processors
~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import re

from elasticapm.utils import six, varmap
from elasticapm.utils.encoding import force_text

MASK = 8 * '*'

SANITIZE_FIELD_NAMES = frozenset([
    'password',
    'secret',
    'passwd',
    'token',
    'api_key',
    'access_token',
    'sessionid',
])

SANITIZE_VALUE_PATTERNS = [
    re.compile(r'^[- \d]{16,19}$'),  # credit card numbers, with or without spacers
]


def remove_http_request_body(client, event):
    """
    Removes request.body from context

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    if 'context' in event and 'request' in event['context']:
        event['context']['request'].pop('body', None)
    return event


def remove_stacktrace_locals(client, event):
    """
    Removes local variables from any frames.

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    func = lambda frame: frame.pop('vars', None)
    return _process_stack_frames(event, func)


def sanitize_stacktrace_locals(client, event):
    """
    Sanitizes local variables in all frames

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    def func(frame):
        if 'vars' in frame:
            frame['vars'] = varmap(_sanitize, frame['vars'])

    return _process_stack_frames(event, func)


def sanitize_http_request_cookies(client, event):
    """
    Sanitizes http request cookies

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """

    # sanitize request.cookies dict
    try:
        cookies = event['context']['request']['cookies']
        event['context']['request']['cookies'] = varmap(_sanitize, cookies)
    except (KeyError, TypeError):
        pass

    # sanitize request.header.cookie string
    try:
        cookie_string = event['context']['request']['headers']['cookie']
        event['context']['request']['headers']['cookie'] = _sanitize_string(
            cookie_string, '; ', '='
        )
    except (KeyError, TypeError):
        pass
    return event


def sanitize_http_request_headers(client, event):
    """
    Sanitizes http request headers

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        headers = event['context']['request']['headers']
        event['context']['request']['headers'] = varmap(_sanitize, headers)
    except (KeyError, TypeError):
        pass
    return event


def sanitize_http_wsgi_env(client, event):
    """
    Sanitizes WSGI environment variables

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        env = event['context']['request']['env']
        event['context']['request']['env'] = varmap(_sanitize, env)
    except (KeyError, TypeError):
        pass
    return event


def sanitize_http_request_querystring(client, event):
    """
    Sanitizes http request query string

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        query_string = force_text(
            event['context']['request']['url']['search'],
            errors='replace'
        )
    except (KeyError, TypeError):
        return event
    if '=' in query_string:
        sanitized_query_string = _sanitize_string(query_string, '&', '=')
        raw = event['context']['request']['url']['raw']
        event['context']['request']['url']['search'] = sanitized_query_string
        event['context']['request']['url']['raw'] = raw.replace(
            query_string,
            sanitized_query_string
        )
    return event


def sanitize_http_request_body(client, event):
    """
    Sanitizes http request body. This only works if the request body
    is a query-encoded string. Other types (e.g. JSON) are not handled by
    this sanitizer.

    :param client: an ElasticAPM client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        body = force_text(
            event['context']['request']['body'],
            errors='replace'
        )
    except (KeyError, TypeError):
        return event
    if '=' in body:
        sanitized_query_string = _sanitize_string(body, '&', '=')
        event['context']['request']['body'] = sanitized_query_string
    return event


def _sanitize(key, value):
    if value is None:
        return

    if isinstance(value, six.string_types) and any(
            pattern.match(value) for pattern in SANITIZE_VALUE_PATTERNS):
        return MASK

    if not key:  # key can be a NoneType
        return value

    key = key.lower()
    for field in SANITIZE_FIELD_NAMES:
        if field in key:
            # store mask as a fixed length for security
            return MASK
    return value


def _sanitize_string(unsanitized, itemsep, kvsep):
    """
    sanitizes a string that contains multiple key/value items
    :param unsanitized: the unsanitized string
    :param itemsep: string that separates items
    :param kvsep: string that separates key from value
    :return: a sanitized string
    """
    sanitized = []
    kvs = unsanitized.split(itemsep)
    for kv in kvs:
        kv = kv.split(kvsep)
        if len(kv) == 2:
            sanitized.append((kv[0], _sanitize(kv[0], kv[1])))
        else:
            sanitized.append(kv)
    return itemsep.join(kvsep.join(kv) for kv in sanitized)


def _process_stack_frames(event, func):
    if 'traces' in event:
        # every trace can have a stack trace
        for trace in event['traces']:
            if 'stacktrace' in trace:
                for frame in trace['stacktrace']:
                    func(frame)
    # an error can have two stacktraces, one in "exception", one in "log"
    if 'exception' in event and 'stacktrace' in event['exception']:
        for frame in event['exception']['stacktrace']:
            func(frame)
    if 'log' in event and 'stacktrace' in event['log']:
        for frame in event['log']['stacktrace']:
            func(frame)
    return event



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace


class RedisInstrumentation(AbstractInstrumentedModule):
    name = 'redis'

    instrument_list = [
        ("redis.client", "Redis.execute_command"),
        ("redis.client", "StrictRedis.execute_command"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        if len(args) > 0:
            wrapped_name = str(args[0])
        else:
            wrapped_name = self.get_wrapped_name(wrapped, instance, method)

        with trace(wrapped_name, "cache.redis", leaf=True):
            return wrapped(*args, **kwargs)


class RedisPipelineInstrumentation(AbstractInstrumentedModule):
    name = 'redis'

    instrument_list = [
        ("redis.client", "BasePipeline.execute"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        wrapped_name = self.get_wrapped_name(wrapped, instance, method)
        with trace(wrapped_name, "cache.redis", leaf=True):
            return wrapped(*args, **kwargs)



from __future__ import absolute_import

import webob

from elasticapm.middleware import ElasticAPM
from tests.utils.compat import TestCase

from ..helpers import get_tempstoreclient


def example_app(environ, start_response):
    raise ValueError('hello world')


class MiddlewareTest(TestCase):
    def setUp(self):
        self.app = example_app

    def test_error_handler(self):
        client = get_tempstoreclient()
        middleware = ElasticAPM(self.app, client=client)

        request = webob.Request.blank('/an-error?foo=bar')
        response = middleware(request.environ, lambda *args: None)

        with self.assertRaises(ValueError):
            list(response)

        self.assertEquals(len(client.events), 1)
        event = client.events.pop(0)['errors'][0]

        self.assertTrue('exception' in event)
        exc = event['exception']
        self.assertEquals(exc['type'], 'ValueError')
        self.assertEquals(exc['message'], 'ValueError: hello world')

        self.assertTrue('request' in event['context'])
        request = event['context']['request']
        self.assertEquals(request['url']['raw'], 'http://localhost/an-error?foo=bar')
        self.assertEquals(request['url']['search'], 'foo=bar')
        self.assertEquals(request['method'], 'GET')
        headers = request['headers']
        self.assertTrue('host' in headers, headers.keys())
        self.assertEquals(headers['host'], 'localhost:80')
        env = request['env']
        self.assertTrue('SERVER_NAME' in env, env.keys())
        self.assertEquals(env['SERVER_NAME'], 'localhost')
        self.assertTrue('SERVER_PORT' in env, env.keys())
        self.assertEquals(env['SERVER_PORT'], '80')



import os
import random
import shutil
import sys
import tempfile

import pytest

from elasticapm.contrib.zerorpc import Middleware
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase

zerorpc = pytest.importorskip("zerorpc")
gevent = pytest.importorskip("gevent")



has_unsupported_pypy = (hasattr(sys, 'pypy_version_info')
                        and sys.pypy_version_info < (2, 6))


class ZeroRPCTest(TestCase):
    def setUp(self):
        self._socket_dir = tempfile.mkdtemp(prefix='elasticapmzerorpcunittest')
        self._server_endpoint = 'ipc://{0}'.format(os.path.join(
                    self._socket_dir, 'random_zeroserver'
        ))

        self._elasticapm_client = get_tempstoreclient()
        zerorpc.Context.get_instance().register_middleware(Middleware(
                    client=self._elasticapm_client
        ))

    @pytest.mark.skipif(has_unsupported_pypy, reason='Failure with pypy < 2.6')
    def test_zerorpc_middleware_with_reqrep(self):
        self._server = zerorpc.Server(random)
        self._server.bind(self._server_endpoint)
        gevent.spawn(self._server.run)

        self._client = zerorpc.Client()
        self._client.connect(self._server_endpoint)

        try:
            self._client.choice([])
        except zerorpc.exceptions.RemoteError as ex:
            self.assertEqual(ex.name, 'IndexError')
            self.assertEqual(len(self._elasticapm_client.events), 1)
            exc = self._elasticapm_client.events[0]['errors'][0]['exception']
            self.assertEqual(exc['type'], 'IndexError')
            frames = exc['stacktrace']
            self.assertEqual(frames[0]['function'], 'choice')
            self.assertEqual(frames[0]['module'], 'random')
            return
        self.fail('An IndexError exception should have been raised an catched')

    def tearDown(self):
        self._client.close()
        self._server.close()
        shutil.rmtree(self._socket_dir, ignore_errors=True)



"""
elasticapm.contrib.django.client
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import

import logging

import django
from django.core.exceptions import DisallowedHost
from django.db import DatabaseError
from django.http import HttpRequest
from django.template import TemplateSyntaxError

from elasticapm.base import Client
from elasticapm.conf import defaults
from elasticapm.contrib.django.utils import (get_data_from_template_debug,
                                             get_data_from_template_source,
                                             iterate_with_template_sources)
from elasticapm.utils import get_url_dict
from elasticapm.utils.wsgi import get_environ, get_headers

try:
    from django.template.loader import LoaderOrigin  # Django < 1.9
except ImportError:
    from django.template.base import Origin as LoaderOrigin  # Django >= 1.9


try:
    # Attempt to use the Django 1.7+ apps sub-framework.
    from django.apps import apps

    is_app_installed = apps.is_installed
except ImportError:
    from django.conf import settings

    # Use the legacy method of simple checking the configuration.
    def is_app_installed(app_name):
        return app_name in settings.INSTALLED_APPS


__all__ = ('DjangoClient',)


class DjangoClient(Client):
    logger = logging.getLogger('elasticapm.errors.client.django')

    def __init__(self, **kwargs):
        instrument_django_middleware = kwargs.pop(
            'instrument_django_middleware',
            None
        )
        if instrument_django_middleware is not None:
            self.instrument_django_middleware = instrument_django_middleware
        else:
            self.instrument_django_middleware = defaults.INSTRUMENT_DJANGO_MIDDLEWARE
        self._framework = 'django'
        self._framework_version = django.get_version()
        super(DjangoClient, self).__init__(**kwargs)

    def get_user_info(self, request):
        user_info = {}

        if not hasattr(request, 'user'):
            return user_info
        try:
            user = request.user
            if hasattr(user, 'is_authenticated'):
                if callable(user.is_authenticated):
                    user_info['is_authenticated'] = user.is_authenticated()
                else:
                    user_info['is_authenticated'] = bool(user.is_authenticated)
            if hasattr(user, 'id'):
                user_info['id'] = user.id
            if hasattr(user, 'get_username'):
                user_info['username'] = user.get_username()
            elif hasattr(user, 'username'):
                user_info['username'] = user.username

            if hasattr(user, 'email'):
                user_info['email'] = user.email
        except DatabaseError:
            # If the connection is closed or similar, we'll just skip this
            return {}

        return user_info

    def get_data_from_request(self, request):
        if request.method != 'GET':
            try:
                if hasattr(request, 'body'):
                    # Django 1.4+
                    raw_data = request.body
                else:
                    raw_data = request.raw_post_data
                data = raw_data if raw_data else request.POST
            except Exception:
                # assume we had a partial read:
                data = '<unavailable>'
        else:
            data = None

        environ = request.META

        result = {
            'body': data,
            'env': dict(get_environ(environ)),
            'headers': dict(get_headers(environ)),
            'method': request.method,
            'socket': {
                'remote_address': request.META.get('REMOTE_ADDR'),
                'encrypted': request.is_secure()
            },
            'cookies': dict(request.COOKIES),
        }

        if hasattr(request, 'get_raw_uri'):
            # added in Django 1.9
            url = request.get_raw_uri()
        else:
            try:
                # Requires host to be in ALLOWED_HOSTS, might throw a
                # DisallowedHost exception
                url = request.build_absolute_uri()
            except DisallowedHost:
                # We can't figure out the real URL, so we have to set it to
                # DisallowedHost
                result['url'] = {'raw': 'DisallowedHost'}
                url = None
        if url:
            result['url'] = get_url_dict(url)
        return result

    def capture(self, event_type, request=None, **kwargs):
        if 'data' not in kwargs:
            kwargs['data'] = data = {}
        else:
            data = kwargs['data']

        if 'context' not in data:
            data['context'] = context = {}
        else:
            context = data['context']

        is_http_request = isinstance(request, HttpRequest)
        if is_http_request:
            context['request'] = self.get_data_from_request(request)
            context['user'] = self.get_user_info(request)

        if kwargs.get('exc_info'):
            exc_value = kwargs['exc_info'][1]
            # As of r16833 (Django) all exceptions may contain a ``django_template_source`` attribute (rather than the
            # legacy ``TemplateSyntaxError.source`` check) which describes template information.
            if hasattr(exc_value, 'django_template_source') or ((isinstance(exc_value, TemplateSyntaxError) and
               isinstance(getattr(exc_value, 'source', None), (tuple, list)) and isinstance(exc_value.source[0], LoaderOrigin))):
                source = getattr(exc_value, 'django_template_source', getattr(exc_value, 'source', None))
                if source is None:
                    self.logger.info('Unable to get template source from exception')
                data.update(get_data_from_template_source(source))
            elif hasattr(exc_value, 'template_debug'):  # Django 1.9+
                data.update(get_data_from_template_debug(exc_value.template_debug))

        result = super(DjangoClient, self).capture(event_type, **kwargs)

        if is_http_request:
            # attach the elasticapm object to the request
            request._elasticapm = {
                'app_name': data.get('app_name', self.app_name),
                'id': self.get_ident(result),
            }

        return result

    def get_stack_info_for_trace(self, frames, extended=True):
        """If the stacktrace originates within the elasticapm module, it will skip
        frames until some other module comes up."""
        frames = list(iterate_with_template_sources(frames, extended))
        i = 0
        while len(frames) > i:
            if 'module' in frames[i] and not (
                    frames[i]['module'].startswith('elasticapm.') or
                    frames[i]['module'] == 'contextlib'
            ):
                return frames[i:]
            i += 1
        return frames

    def send(self, **kwargs):
        """
        Serializes and signs ``data`` and passes the payload off to ``send_remote``

        If ``servers`` was passed into the constructor, this will serialize the data and pipe it to
        each server using ``send_remote()``.
        """
        if self.servers:
            return super(DjangoClient, self).send(**kwargs)
        else:
            self.error_logger.error('No servers configured, and elasticapm not installed. Cannot send message')
            return None



"""
elasticapm.contrib.django.celery.models
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

if 'djcelery' not in settings.INSTALLED_APPS:
    raise ImproperlyConfigured("Put 'djcelery' in your "
        "INSTALLED_APPS setting in order to use the sentry celery client.")



# -*- coding: utf-8 -*-
import platform
import time

import mock
import pytest

import elasticapm
from elasticapm.base import Client, ClientState
from elasticapm.conf import defaults
from elasticapm.transport.base import Transport, TransportException
from elasticapm.transport.http import HTTPTransport
from elasticapm.utils import six
from elasticapm.utils.compat import urlparse
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class ClientStateTest(TestCase):
    def test_should_try_online(self):
        state = ClientState()
        self.assertEquals(state.should_try(), True)

    def test_should_try_new_error(self):
        state = ClientState()
        state.status = state.ERROR
        state.last_check = time.time()
        state.retry_number = 1
        self.assertEquals(state.should_try(), False)

    def test_should_try_time_passed_error(self):
        state = ClientState()
        state.status = state.ERROR
        state.last_check = time.time() - 10
        state.retry_number = 1
        self.assertEquals(state.should_try(), True)

    def test_set_fail(self):
        state = ClientState()
        state.set_fail()
        self.assertEquals(state.status, state.ERROR)
        self.assertNotEquals(state.last_check, None)
        self.assertEquals(state.retry_number, 1)

    def test_set_success(self):
        state = ClientState()
        state.status = state.ERROR
        state.last_check = 'foo'
        state.retry_number = 0
        state.set_success()
        self.assertEquals(state.status, state.ONLINE)
        self.assertEquals(state.last_check, None)
        self.assertEquals(state.retry_number, 0)


class DummyTransport(Transport):
    def send(self, data, headers):
        pass


class ClientTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()

    def test_app_info(self):
        app_info = self.client.get_app_info()
        assert app_info['name'] == self.client.app_name
        assert app_info['language'] == {
            'name': 'python',
            'version': platform.python_version()
        }

    def test_system_info(self):
        system_info = self.client.get_system_info()
        assert {'hostname', 'architecture', 'platform'} == set(system_info.keys())

    def test_config_by_environment(self):
        with mock.patch.dict('os.environ', {
            'ELASTICAPM_APP_NAME': 'app',
            'ELASTICAPM_SECRET_TOKEN': 'token',
            'ELASTICAPM_GIT_REF': 'aabbccdd'
        }):
            client = Client()
            self.assertEqual(client.app_name, 'app')
            self.assertEqual(client.secret_token, 'token')
            self.assertEqual(client.git_ref, 'aabbccdd')
            self.assertEqual(client.is_send_disabled, False)
        with mock.patch.dict('os.environ', {
            'ELASTICAPM_DISABLE_SEND': 'true',
        }):
            client = Client()
            self.assertEqual(client.is_send_disabled, True)

    def test_config_non_string_types(self):
        """
        tests if we can handle non string types as configuration, e.g.
        Value types from django-configuration
        """
        class MyValue(object):
            def __init__(self, content):
                self.content = content

            def __str__(self):
                return str(self.content)

            def __repr__(self):
                return repr(self.content)

        client = Client(
            servers=['localhost'],
            app_name=MyValue('bar'),
            secret_token=MyValue('bay')
        )
        assert isinstance(client.secret_token, six.string_types)
        assert isinstance(client.app_name, six.string_types)

    def test_custom_transport(self):
        client = Client(
            servers=['localhost'],
            app_name='bar',
            secret_token='baz',
            transport_class='tests.client.client_tests.DummyTransport',
        )
        self.assertEqual(client._transport_class, DummyTransport)

    def test_empty_processor_list(self):
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            processors=[],
        )

        self.assertEqual(client.processors, [])

    @mock.patch('elasticapm.transport.http_urllib3.Urllib3Transport.send')
    @mock.patch('elasticapm.base.ClientState.should_try')
    def test_send_remote_failover_sync(self, should_try, http_send):
        should_try.return_value = True

        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            async_mode=False,
        )
        logger = mock.Mock()
        client.error_logger.error = logger

        # test error
        encoded_data = client.encode({'message': 'oh no'})
        http_send.side_effect = TransportException('oopsie', encoded_data)
        client.send_remote('http://example.com/api/store', data=encoded_data)
        assert client.state.status == client.state.ERROR
        assert len(logger.call_args_list) == 2
        assert 'oopsie' in logger.call_args_list[0][0][0]
        assert 'oh no' in logger.call_args_list[1][0][1]

        # test recovery
        http_send.side_effect = None
        client.send_remote('http://example.com/api/store', 'foo')
        assert client.state.status == client.state.ONLINE

    @mock.patch('elasticapm.transport.http_urllib3.Urllib3Transport.send')
    @mock.patch('elasticapm.base.ClientState.should_try')
    def test_send_remote_failover_sync_stdlib(self, should_try, http_send):
        should_try.return_value = True

        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            async_mode=False,
        )
        logger = mock.Mock()
        client.error_logger.error = logger

        # test error
        encoded_data = client.encode({'message': 'oh no'})
        http_send.side_effect = ValueError('oopsie')
        client.send_remote('http://example.com/api/store', data=encoded_data)
        assert client.state.status == client.state.ERROR
        assert len(logger.call_args_list) == 1
        assert 'oopsie' in logger.call_args_list[0][0][1]

        # test recovery
        http_send.side_effect = None
        client.send_remote('http://example.com/api/store', 'foo')
        assert client.state.status == client.state.ONLINE

    @mock.patch('elasticapm.transport.http_urllib3.Urllib3Transport.send')
    @mock.patch('elasticapm.base.ClientState.should_try')
    def test_send_remote_failover_async(self, should_try, http_send):
        should_try.return_value = True

        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            async_mode=True,
        )
        logger = mock.Mock()
        client.error_logger.error = logger

        # test error
        encoded_data = client.encode({'message': 'oh no'})
        http_send.side_effect = TransportException('oopsie', encoded_data)
        client.send_remote('http://example.com/api/store', data=encoded_data)
        client.close()
        assert client.state.status == client.state.ERROR
        assert len(logger.call_args_list) == 2
        assert 'oopsie' in logger.call_args_list[0][0][0]
        assert 'oh no' in logger.call_args_list[1][0][1]

        # test recovery
        http_send.side_effect = None
        client.send_remote('http://example.com/api/store', 'foo')
        client.close()
        assert client.state.status == client.state.ONLINE

    @mock.patch('elasticapm.base.Client.send_remote')
    @mock.patch('elasticapm.base.time.time')
    def test_send(self, time, send_remote):
        time.return_value = 1328055286.51
        public = "public"
        access_token = "secret"
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
        )
        client.send(**{
            'foo': 'bar',
        })
        send_remote.assert_called_once_with(
            url='http://example.com',
            data=six.b('x\x9c\xabVJ\xcb\xcfW\xb2RPJJ,R\xaa\x05\x00 \x98\x04T'),
            headers={
                'Content-Type': 'application/json',
                'Content-Encoding': 'deflate',
                'Authorization': 'Bearer %s' % (access_token),
                'User-Agent': 'elasticapm-python/%s' % elasticapm.VERSION,
            },
        )

    @mock.patch('elasticapm.base.Client.send_remote')
    @mock.patch('elasticapm.base.time.time')
    def test_send_not_enabled(self, time, send_remote):
        time.return_value = 1328055286.51
        with mock.patch.dict('os.environ', {'ELASTICAPM_DISABLE_SEND': 'true'}):
            client = Client(
                servers=['http://example.com'],
                app_name='app_name',
                secret_token='secret',
            )
        client.send(**{
            'foo': 'bar',
        })

        assert not send_remote.called

    @mock.patch('elasticapm.base.Client.send_remote')
    @mock.patch('elasticapm.base.time.time')
    def test_send_with_auth_header(self, time, send_remote):
        time.return_value = 1328055286.51
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
        )
        client.send(auth_header='foo', **{
            'foo': 'bar',
        })
        send_remote.assert_called_once_with(
            url='http://example.com',
            data=six.b('x\x9c\xabVJ\xcb\xcfW\xb2RPJJ,R\xaa\x05\x00 \x98\x04T'),
            headers={
                'Content-Type': 'application/json',
                'Content-Encoding': 'deflate',
                'Authorization': 'foo',
                'User-Agent': 'elasticapm-python/%s' % elasticapm.VERSION,
            },
        )

    @mock.patch('elasticapm.transport.http_urllib3.Urllib3Transport.send')
    @mock.patch('elasticapm.transport.http_urllib3.Urllib3Transport.close')
    @mock.patch('elasticapm.base.Client._traces_collect')
    def test_client_shutdown_sync(self, mock_traces_collect, mock_close,
                                  mock_send):
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            async_mode=False,
        )
        client.send(auth_header='foo', **{
            'foo': 'bar',
        })
        client.close()
        self.assertEqual(mock_close.call_count, 1)
        self.assertEqual(mock_traces_collect.call_count, 1)

    @mock.patch('elasticapm.transport.http_urllib3.Urllib3Transport.send')
    @mock.patch('elasticapm.base.Client._traces_collect')
    def test_client_shutdown_async(self, mock_traces_collect, mock_send):
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            async_mode=True,
        )
        client.send(auth_header='foo', **{
            'foo': 'bar',
        })
        client.close()
        self.assertEqual(mock_traces_collect.call_count, 1)
        self.assertEqual(mock_send.call_count, 1)

    def test_encode_decode(self):
        data = {'foo': 'bar'}
        encoded = self.client.encode(data)
        self.assertTrue(type(encoded), str)
        self.assertEquals(data, self.client.decode(encoded))

    def test_explicit_message_on_message_event(self):
        self.client.capture('Message', message='test', data={
            'message': 'foo'
        })

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['message'], 'foo')

    def test_explicit_message_on_exception_event(self):
        try:
            raise ValueError('foo')
        except:
            self.client.capture('Exception', data={'message': 'foobar'})

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['message'], 'foobar')

    def test_exception_event(self):
        try:
            raise ValueError('foo')
        except:
            self.client.capture('Exception')

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertTrue('exception' in event)
        exc = event['exception']
        self.assertEquals(exc['message'], 'ValueError: foo')
        self.assertEquals(exc['type'], 'ValueError')
        self.assertEquals(exc['module'], ValueError.__module__)  # this differs in some Python versions
        self.assertTrue('stacktrace' in exc)
        frames = exc['stacktrace']
        self.assertEquals(len(frames), 1)
        frame = frames[0]
        self.assertEquals(frame['abs_path'], __file__.replace('.pyc', '.py'))
        self.assertEquals(frame['filename'], 'tests/client/client_tests.py')
        self.assertEquals(frame['module'], __name__)
        self.assertEquals(frame['function'], 'test_exception_event')
        self.assertTrue('timestamp' in event)

    def test_message_event(self):
        self.client.capture('Message', message='test')

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['log']['message'], 'test')
        self.assertFalse('stacktrace' in event)
        self.assertTrue('timestamp' in event)

    # def test_stack_explicit_frames(self):
    #     def bar():
    #         return inspect.stack()

    #     frames = bar()

    #     self.client.capture('test', stack=iter_stack_frames(frames))

    #     self.assertEquals(len(self.client.events), 1)
    #     event = self.client.events.pop(0)
    #     self.assertEquals(event['message'], 'test')
    #     self.assertTrue('stacktrace' in event)
    #     self.assertEquals(len(frames), len(event['stacktrace']['frames']))
    #     for frame, frame_i in zip(frames, event['stacktrace']['frames']):
    #         self.assertEquals(frame[0].f_code.co_filename, frame_i['abs_path'])
    #         self.assertEquals(frame[0].f_code.co_name, frame_i['function'])

    # def test_stack_auto_frames(self):
    #     self.client.create_from_text('test', stack=True)

    #     self.assertEquals(len(self.client.events), 1)
    #     event = self.client.events.pop(0)
    #     self.assertEquals(event['message'], 'test')
    #     self.assertTrue('stacktrace' in event)
    #     self.assertTrue('timestamp' in event)

    # def test_site(self):
    #     self.client.capture('Message', message='test', data={'site': 'test'})

    #     self.assertEquals(len(self.client.events), 1)
    #     event = self.client.events.pop(0)
    #     self.assertEquals(event['site'], 'test')
    #     self.assertTrue('timestamp' in event)

    # def test_implicit_site(self):
    #     self.client = TempStoreClient(site='foo')
    #     self.client.capture('Message', message='test')

    #     self.assertEquals(len(self.client.events), 1)
    #     event = self.client.events.pop(0)
    #     self.assertEquals(event['site'], 'foo')

    def test_logger(self):
        self.client.capture('Message', message='test', data={'logger': 'test'})

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['logger'], 'test')
        self.assertTrue('timestamp' in event)

    @mock.patch('elasticapm.base.Client.send')
    @mock.patch('elasticapm.base.TransactionsStore.should_collect')
    def test_metrics_collection(self, should_collect, mock_send):
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
        )
        should_collect.return_value = False
        for i in range(7):
            client.begin_transaction("transaction.test")
            client.end_transaction('test-transaction', 200)

        self.assertEqual(len(client.instrumentation_store), 7)
        self.assertEqual(mock_send.call_count, 0)
        should_collect.return_value = True

        client.begin_transaction("transaction.test")
        client.end_transaction('my-other-transaction', 200)
        self.assertEqual(len(client.instrumentation_store), 0)
        self.assertEqual(mock_send.call_count, 1)

    @mock.patch('elasticapm.base.Client.send')
    @mock.patch('elasticapm.base.TransactionsStore.should_collect')
    def test_call_end_twice(self, should_collect, mock_send):
        client = get_tempstoreclient()

        should_collect.return_value = False
        client.begin_transaction("celery")

        client.end_transaction('test-transaction', 200)
        client.end_transaction('test-transaction', 200)

    @mock.patch('elasticapm.utils.is_master_process')
    def test_client_uses_sync_mode_when_master_process(self, is_master_process):
        # when in the master process, the client should use the non-async
        # HTTP transport, even if async_mode is True
        is_master_process.return_value = True
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            async_mode=True,
        )
        self.assertIsInstance(
            client._get_transport(urlparse.urlparse('http://exampe.com')),
            HTTPTransport
        )

    @mock.patch('elasticapm.base.Client.send')
    @mock.patch('elasticapm.base.TransactionsStore.should_collect')
    def test_ignore_patterns(self, should_collect, mock_send):
        client = Client(
            servers=['http://example.com'],
            app_name='app_name',
            secret_token='secret',
            async_mode=True,
            transactions_ignore_patterns=[
                '^OPTIONS',
                'views.api.v2'
            ]
        )

        should_collect.return_value = False
        client.begin_transaction("web")
        client.end_transaction('OPTIONS views.healthcheck', 200)

        client.begin_transaction("web")
        client.end_transaction('GET views.users', 200)

        self.assertEqual(len(client.instrumentation_store), 1)



import random
import threading

import mock
import urllib3

import elasticapm
import elasticapm.instrumentation.control
from elasticapm.traces import trace
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase

try:
    from http import server as SimpleHTTPServer
    from socketserver import TCPServer
except ImportError:
    import SimpleHTTPServer
    from SocketServer import TCPServer

class MyTCPServer(TCPServer):
    allow_reuse_address = True

class InstrumentUrllib3Test(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        self.port = random.randint(50000, 60000)
        self.start_test_server()
        elasticapm.instrumentation.control.instrument()

    def tearDown(self):
        if self.httpd:
            self.httpd.shutdown()

    def start_test_server(self):
        handler = SimpleHTTPServer.SimpleHTTPRequestHandler

        self.httpd = MyTCPServer(("", self.port), handler)

        self.httpd_thread = threading.Thread(target=self.httpd.serve_forever)
        self.httpd_thread.setDaemon(True)
        self.httpd_thread.start()

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_urllib3(self, should_collect):
        should_collect.return_value = False
        self.client.begin_transaction("transaction")
        expected_sig = 'GET localhost:{0}'.format(self.port)
        with trace("test_pipeline", "test"):
            pool = urllib3.PoolManager(timeout=0.1)

            url = 'http://localhost:{0}/hello_world'.format(self.port)
            r = pool.request('GET', url)

        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', 'test_pipeline', expected_sig]

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(len(traces), 3)

        self.assertEqual(traces[0]['name'], expected_sig)
        self.assertEqual(traces[0]['type'], 'ext.http.urllib3')
        self.assertEqual(traces[0]['context']['url'], url)
        self.assertEqual(traces[0]['parent'], 1)

        self.assertEqual(traces[1]['name'], 'test_pipeline')
        self.assertEqual(traces[1]['type'], 'test')

        self.assertEqual(traces[2]['name'], 'transaction')
        self.assertEqual(traces[2]['type'], 'transaction')



import asyncio

import aiohttp

from elasticapm.conf import defaults

from .base import TransportException
from .http import HTTPTransport


class AsyncioHTTPTransport(HTTPTransport):
    """
    HTTP Transport ready for asyncio
    """

    def __init__(self, parsed_url):
        self.check_scheme(parsed_url)

        self._parsed_url = parsed_url
        self._url = parsed_url.geturl()
        loop = asyncio.get_event_loop()
        self.client = aiohttp.ClientSession(loop=loop)

    async def send(self, data, headers, timeout=None):
        """Use synchronous interface, because this is a coroutine."""

        if timeout is None:
            timeout = defaults.TIMEOUT
        try:
            with aiohttp.Timeout(timeout):
                async with self.client.post(self._url,
                                            data=data,
                                            headers=headers) as response:
                    assert response.status == 202
        except asyncio.TimeoutError as e:
            print_trace = True
            message = ("Connection to APM Server timed out "
                       "(url: %s, timeout: %d seconds)" % (self._url, timeout))
            raise TransportException(message, data,
                                     print_trace=print_trace) from e
        except AssertionError as e:
            print_trace = True
            body = await response.read()
            if response.status == 429:
                message = 'Temporarily rate limited: '
                print_trace = False
            else:
                message = 'Unable to reach APM Server: '
            message += '%s (url: %s, body: %s)' % (e, self._url, body)
            raise TransportException(message, data,
                                     print_trace=print_trace) from e
        except Exception as e:
            print_trace = True
            message = 'Unable to reach APM Server: %s (url: %s)' % (
                e, self._url)
            raise TransportException(message, data,
                                     print_trace=print_trace) from e
        else:
            return response.headers.get('Location')

    def __del__(self):
        self.client.close()



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace


class DjangoTemplateInstrumentation(AbstractInstrumentedModule):
    name = 'django_template'

    instrument_list = [
        ("django.template", "Template.render"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        name = getattr(instance, 'name', None)

        if not name:
            name = '<template string>'
        with trace(name, "template.django"):
            return wrapped(*args, **kwargs)


class DjangoTemplateSourceInstrumentation(AbstractInstrumentedModule):
    name = 'django_template_source'
    instrument_list = [
        ('django.template.base', 'Parser.extend_nodelist')
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        ret = wrapped(*args, **kwargs)

        if len(args) > 1:
            node = args[1]
        elif 'node' in kwargs:
            node = kwargs['node']
        else:
            return ret

        if len(args) > 2:
            token = args[2]
        elif 'token' in kwargs:
            token = kwargs['token']
        else:
            return ret

        if not hasattr(node, 'token') and hasattr(token, 'lineno'):
            node.token = token

        return ret



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django import VERSION as DJANGO_VERSION
from django.db import models

if DJANGO_VERSION >= (1, 5):
    from django.contrib.auth.models import AbstractBaseUser, BaseUserManager

    class MyUser(AbstractBaseUser):
        USERNAME_FIELD = 'my_username'
        my_username = models.CharField(max_length=30)

        objects = BaseUserManager()

        class Meta:
            abstract = False



"""
Backported LRU cache from Python 3.3
http://code.activestate.com/recipes/578078-py26-and-py30-backport-of-python-33s-lru-cache/

"""
from collections import namedtuple
from threading import RLock

_CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize", "currsize"])


class LRUCache(object):
    def __init__(self, maxsize=100):
        self.cache = dict()
        self.cache_get = self.cache.get           # bound method to lookup key or return None
        self._len = len                      # localize the global len() function
        self.lock = RLock()                  # because linkedlist updates aren't threadsafe
        self.root = []                       # root of the circular doubly linked list
        self.root[:] = [self.root, self.root, None]      # initialize by pointing to self
        self.nonlocal_root = [self.root]                  # make updateable non-locally
        self.PREV, self.NEXT, self.KEY = 0, 1, 2    # names for the link fields
        self.maxsize = maxsize

    def __contains__(self, key):
        with self.lock:
            link = self.cache_get(key)
            if link is not None:
                # record recent use of the key by moving it to the front of the list
                root, = self.nonlocal_root
                link_prev, link_next, key = link
                link_prev[self.NEXT] = link_next
                link_next[self.PREV] = link_prev
                last = root[self.PREV]
                last[self.NEXT] = root[self.PREV] = link
                link[self.PREV] = last
                link[self.NEXT] = root

                return True
        return False

    def set(self, key):
        with self.lock:
            root, = self.nonlocal_root
            if key in self.cache:
                # getting here means that this same key was added to the
                # cache while the lock was released.  since the link
                # update is already done, we need only return the
                # computed result and update the count of misses.
                pass
            elif self._len(self.cache) >= self.maxsize:
                # use the old root to store the new key and result
                oldroot = root
                oldroot[self.KEY] = key
                # empty the oldest link and make it the new root
                root = self.nonlocal_root[0] = oldroot[self.NEXT]
                oldkey = root[self.KEY]
                root[self.KEY] = None
                # now update the cache dictionary for the new links
                del self.cache[oldkey]
                self.cache[key] = oldroot
            else:
                # put result in a new link at the front of the list
                last = root[self.PREV]
                link = [last, root, key]
                last[self.NEXT] = root[self.PREV] = self.cache[key] = link



import logbook

from elasticapm.handlers.logbook import LogbookHandler
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class LogbookHandlerTest(TestCase):
    def setUp(self):
        self.logger = logbook.Logger(__name__)
        self.client = get_tempstoreclient(include_paths=['tests', 'elasticapm'])
        self.handler = LogbookHandler(self.client)

    def test_logger_error_level(self):
        with self.handler.applicationbound():
            self.logger.error('This is a test error')

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        assert event['log']['logger_name'] == __name__
        assert event['log']['level'] == "error"
        assert event['log']['message'] == 'This is a test error'
        self.assertFalse('stacktrace' in event['log'])
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        assert event['log']['param_message'] == 'This is a test error'

    def test_logger_warning_level(self):
        with self.handler.applicationbound():
            self.logger.warning('This is a test warning')
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        assert event['log']['logger_name'] == __name__
        assert event['log']['level'] == "warning"
        assert event['log']['message'] == 'This is a test warning'
        self.assertFalse('stacktrace' in event['log'])
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        assert event['log']['param_message'] == 'This is a test warning'

    def test_logger_with_extra(self):
        with self.handler.applicationbound():
            self.logger.info('This is a test info with a url', extra=dict(
                url='http://example.com',
            ))
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['context']['custom']['url'], 'http://example.com')
        self.assertFalse('stacktrace' in event['log'])
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test info with a url')

    def test_logger_with_exc_info(self):
        with self.handler.applicationbound():
            try:
                raise ValueError('This is a test ValueError')
            except ValueError:
                self.logger.info('This is a test info with an exception', exc_info=True)

        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]

        self.assertEquals(event['log']['message'], 'This is a test info with an exception')
        self.assertFalse('stacktrace' in event['log'])
        self.assertTrue('exception' in event)
        exc = event['exception']
        self.assertEquals(exc['type'], 'ValueError')
        self.assertEquals(exc['message'], 'ValueError: This is a test ValueError')
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test info with an exception')

    def test_logger_param_message(self):
        with self.handler.applicationbound():
            self.logger.info('This is a test of %s', 'args')
        self.assertEquals(len(self.client.events), 1)
        event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(event['log']['message'], 'This is a test of args')
        self.assertFalse('stacktrace' in event['log'])
        self.assertFalse('exception' in event)
        self.assertTrue('param_message' in event['log'])
        self.assertEquals(event['log']['param_message'], 'This is a test of %s')

    def test_client_arg(self):
        client = get_tempstoreclient(include_paths=['tests'])
        handler = LogbookHandler(client)
        self.assertEquals(handler.client, client)

    def test_client_kwarg(self):
        client = get_tempstoreclient(include_paths=['tests'])
        handler = LogbookHandler(client=client)
        self.assertEquals(handler.client, client)

    def test_invalid_first_arg_type(self):
        self.assertRaises(ValueError, LogbookHandler, object)

    def test_missing_client_arg(self):
        self.assertRaises(TypeError, LogbookHandler)



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace
from elasticapm.utils import default_ports
from elasticapm.utils.compat import urlparse


def get_host_from_url(url):
    parsed_url = urlparse.urlparse(url)
    host = parsed_url.hostname or " "

    if (
        parsed_url.port and
        default_ports.get(parsed_url.scheme) != parsed_url.port
    ):
        host += ":" + str(parsed_url.port)

    return host


class RequestsInstrumentation(AbstractInstrumentedModule):
    name = 'requests'

    instrument_list = [
        ("requests.sessions", "Session.send"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        if 'request' in kwargs:
            request = kwargs['request']
        else:
            request = args[0]

        signature = request.method.upper()
        signature += " " + get_host_from_url(request.url)

        with trace(signature, "ext.http.requests",
                   {'url': request.url}, leaf=True):
            return wrapped(*args, **kwargs)



from elasticapm.utils.module_import import import_string

_cls_register = set([
    'elasticapm.instrumentation.packages.botocore.BotocoreInstrumentation',
    'elasticapm.instrumentation.packages.jinja2.Jinja2Instrumentation',
    'elasticapm.instrumentation.packages.psycopg2.Psycopg2Instrumentation',
    'elasticapm.instrumentation.packages.psycopg2.Psycopg2RegisterTypeInstrumentation',
    'elasticapm.instrumentation.packages.mysql.MySQLInstrumentation',
    'elasticapm.instrumentation.packages.pylibmc.PyLibMcInstrumentation',
    'elasticapm.instrumentation.packages.pymongo.PyMongoInstrumentation',
    'elasticapm.instrumentation.packages.pymongo.PyMongoBulkInstrumentation',
    'elasticapm.instrumentation.packages.pymongo.PyMongoCursorInstrumentation',
    'elasticapm.instrumentation.packages.python_memcached.PythonMemcachedInstrumentation',
    'elasticapm.instrumentation.packages.redis.RedisInstrumentation',
    'elasticapm.instrumentation.packages.redis.RedisPipelineInstrumentation',
    'elasticapm.instrumentation.packages.requests.RequestsInstrumentation',
    'elasticapm.instrumentation.packages.sqlite.SQLiteInstrumentation',
    'elasticapm.instrumentation.packages.urllib3.Urllib3Instrumentation',

    'elasticapm.instrumentation.packages.django.template.DjangoTemplateInstrumentation',
    'elasticapm.instrumentation.packages.django.template.DjangoTemplateSourceInstrumentation',
])


def register(cls):
    _cls_register.add(cls)

_instrumentation_singletons = {}


def get_instrumentation_objects():
    for cls_str in _cls_register:
        if cls_str not in _instrumentation_singletons:
            cls = import_string(cls_str)
            _instrumentation_singletons[cls_str] = cls()

        obj = _instrumentation_singletons[cls_str]
        yield obj



import os

import mock
from jinja2 import Environment, FileSystemLoader
from jinja2.environment import Template

import elasticapm.instrumentation.control
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class InstrumentJinja2Test(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        filedir = os.path.dirname(__file__)
        loader = FileSystemLoader(filedir)
        self.env = Environment(loader=loader)
        elasticapm.instrumentation.control.instrument()

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_from_file(self, should_collect):
        should_collect.return_value = False
        self.client.begin_transaction("transaction.test")
        template = self.env.get_template('mytemplate.html')
        template.render()
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', 'mytemplate.html']

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(traces[0]['name'], 'mytemplate.html')
        self.assertEqual(traces[0]['type'], 'template.jinja2')

    def test_from_string(self):
        self.client.begin_transaction("transaction.test")
        template = Template("<html></html")
        template.render()
        self.client.end_transaction("test")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', '<template>']

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(traces[0]['name'], '<template>')
        self.assertEqual(traces[0]['type'], 'template.jinja2')



from __future__ import absolute_import

from elasticapm.utils.lru import LRUCache
from tests.utils.compat import TestCase


class LRUTest(TestCase):
    def test_insert_overflow(self):

        lru = LRUCache(4)

        for x in range(6):
            lru.set(x)

        self.assertNotIn(1, lru)
        for x in range(2, 6):
            self.assertIn(x, lru)



from __future__ import absolute_import

import django
from django.conf import settings
from django.conf.urls import url
from django.http import HttpResponse

from tests.contrib.django.testapp import views


def handler500(request):
    if getattr(settings, 'BREAK_THAT_500', False):
        raise ValueError('handler500')
    return HttpResponse('')


urlpatterns = (
    url(r'^render-heavy-template$', views.render_template_view, name='render-heavy-template'),
    url(r'^render-user-template$', views.render_user_view, name='render-user-template'),
    url(r'^no-error$', views.no_error, name='elasticapm-no-error'),
    url(r'^no-error-slash/$', views.no_error, name='elasticapm-no-error-slash'),
    url(r'^ignored-exception/$', views.ignored_exception, name='elasticapm-ignored-exception'),
    url(r'^fake-login$', views.fake_login, name='elasticapm-fake-login'),
    url(r'^trigger-500$', views.raise_exc, name='elasticapm-raise-exc'),
    url(r'^trigger-500-ioerror$', views.raise_ioerror, name='elasticapm-raise-ioerror'),
    url(r'^trigger-500-decorated$', views.decorated_raise_exc, name='elasticapm-raise-exc-decor'),
    url(r'^trigger-500-django$', views.django_exc, name='elasticapm-django-exc'),
    url(r'^trigger-500-template$', views.template_exc, name='elasticapm-template-exc'),
    url(r'^trigger-500-log-request$', views.logging_request_exc, name='elasticapm-log-request-exc'),
)


if django.VERSION >= (1, 8):
    urlpatterns += url(r'^render-jinja2-template$', views.render_jinja2_template,
        name='render-jinja2-template'),



from __future__ import absolute_import

from twisted.python.failure import Failure

from elasticapm.contrib.twisted import LogObserver
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class TwistedLogObserverTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()

    def test_observer(self):
        observer = LogObserver(client=self.client)
        try:
            1 / 0
        except ZeroDivisionError:
            failure = Failure()
        event = dict(log_failure=failure)
        observer(event)

        cli_event = self.client.events.pop(0)['errors'][0]
        self.assertEquals(cli_event['exception']['type'], 'ZeroDivisionError')



# -*- coding: utf-8 -*-
import os

import pytest

from elasticapm.instrumentation import control
from elasticapm.instrumentation.packages.psycopg2 import (PGCursorProxy,
                                                          extract_signature)
from tests.contrib.django.django_tests import get_client

try:
    import psycopg2
    has_psycopg2 = True
except ImportError:
    has_psycopg2 = False

travis_and_psycopg2 = 'TRAVIS' not in os.environ or not has_psycopg2


@pytest.yield_fixture(scope='function')
def postgres_connection(request):
    conn = psycopg2.connect(
        database=os.environ.get('POSTGRES_DB', 'elasticapm_test'),
        user=os.environ.get('POSTGRES_USER', 'postgres'),
        host=os.environ.get('POSTGRES_HOST', None),
        port=os.environ.get('POSTGRES_PORT', None),
    )
    cursor = conn.cursor()
    cursor.execute(
        "CREATE TABLE test(id int, name VARCHAR(5) NOT NULL);"
        "INSERT INTO test VALUES (1, 'one'), (2, 'two'), (3, 'three');"
    )

    yield conn

    # cleanup
    cursor.execute('ROLLBACK')


def test_insert():
    sql = """INSERT INTO mytable (id, name) VALUE ('2323', 'Ron')"""
    actual = extract_signature(sql)

    assert "INSERT INTO mytable" == actual


def test_update_with_quotes():
    sql = """UPDATE "my table" set name='Ron' WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "UPDATE my table" == actual


def test_update():
    sql = """update mytable set name = 'Ron where id = 'a'"""
    actual = extract_signature(sql)

    assert "UPDATE mytable" == actual


def test_delete_simple():
    sql = 'DELETE FROM "mytable"'
    actual = extract_signature(sql)

    assert "DELETE FROM mytable" == actual


def test_delete():
    sql = """DELETE FROM "my table" WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "DELETE FROM my table" == actual


def test_select_simple():
    sql = """SELECT id, name FROM my_table WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM my_table" == actual


def test_select_with_entity_quotes():
    sql = """SELECT id, name FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_difficult_values():
    sql = """SELECT id, 'some name' + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_dollar_quotes():
    sql = """SELECT id, $$some single doubles ' $$ + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_invalid_dollar_quotes():
    sql = """SELECT id, $fish$some single doubles ' $$ + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM" == actual


def test_select_with_dollar_quotes_custom_token():
    sql = """SELECT id, $token $FROM $ FROM $ FROM single doubles ' $token $ + '" from Denmark' FROM "mytable" WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_difficult_table_name():
    sql = "SELECT id FROM \"myta\n-æøåble\" WHERE id = 2323"""
    actual = extract_signature(sql)

    assert "SELECT FROM myta\n-æøåble" == actual


def test_select_subselect():
    sql = """SELECT id, name FROM (
            SELECT id, 'not a FROM ''value' FROM mytable WHERE id = 2323
    ) LIMIT 20"""
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_subselect_with_alias():
    sql = """
    SELECT count(*)
    FROM (
        SELECT count(id) AS some_alias, some_column
        FROM mytable
        GROUP BY some_colun
        HAVING count(id) > 1
    ) AS foo
    """
    actual = extract_signature(sql)

    assert "SELECT FROM mytable" == actual


def test_select_with_multiple_tables():
    sql = """SELECT count(table2.id)
        FROM table1, table2, table2
        WHERE table2.id = table1.table2_id
    """
    actual = extract_signature(sql)
    assert "SELECT FROM table1" == actual


def test_select_with_invalid_subselect():
    sql = "SELECT id FROM (SELECT * """
    actual = extract_signature(sql)

    assert "SELECT FROM" == actual


def test_select_with_invalid_literal():
    sql = "SELECT 'neverending literal FROM (SELECT * FROM ..."""
    actual = extract_signature(sql)

    assert "SELECT FROM" == actual


def test_savepoint():
    sql = """SAVEPOINT x_asd1234"""
    actual = extract_signature(sql)

    assert "SAVEPOINT" == actual


def test_begin():
    sql = """BEGIN"""
    actual = extract_signature(sql)

    assert "BEGIN" == actual


def test_create_index_with_name():
    sql = """CREATE INDEX myindex ON mytable"""
    actual = extract_signature(sql)

    assert "CREATE INDEX" == actual


def test_create_index_without_name():
    sql = """CREATE INDEX ON mytable"""
    actual = extract_signature(sql)

    assert "CREATE INDEX" == actual


def test_drop_table():
    sql = """DROP TABLE mytable"""
    actual = extract_signature(sql)

    assert "DROP TABLE" == actual


def test_multi_statement_sql():
    sql = """CREATE TABLE mytable; SELECT * FROM mytable; DROP TABLE mytable"""
    actual = extract_signature(sql)

    assert "CREATE TABLE" == actual

@pytest.mark.skipif(travis_and_psycopg2,
                    reason="Requires postgres server. Only runs on travisci.")
def test_psycopg2_register_type(postgres_connection):
    import psycopg2.extras

    client = get_client()
    control.instrument()

    try:
        client.begin_transaction("web.django")
        new_type = psycopg2.extras.register_uuid(None, postgres_connection)
        client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the traces for the other tests.
        client.instrumentation_store.get_all()

    assert new_type is not None


@pytest.mark.skipif(travis_and_psycopg2,
                    reason="Requires postgres server. Only runs on travisci.")
def test_psycopg2_register_json(postgres_connection):
    # register_json bypasses register_type, so we have to test unwrapping
    # separately
    import psycopg2.extras

    client = get_client()
    control.instrument()

    try:
        client.begin_transaction("web.django")
        # as arg
        new_type = psycopg2.extras.register_json(postgres_connection,
                                                 loads=lambda x: x)
        assert new_type is not None
        # as kwarg
        new_type = psycopg2.extras.register_json(conn_or_curs=postgres_connection,
                                                 loads=lambda x: x)
        assert new_type is not None
        client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the traces for the other tests.
        client.instrumentation_store.get_all()


@pytest.mark.skipif(travis_and_psycopg2,
                    reason="Requires postgres server. Only runs on travisci.")
def test_psycopg2_tracing_outside_of_elasticapm_transaction(postgres_connection):
    client = get_client()
    control.instrument()
    cursor = postgres_connection.cursor()
    # check that the cursor is a proxy, even though we're not in an elasticapm
    # transaction
    assert isinstance(cursor, PGCursorProxy)
    cursor.execute('SELECT 1')
    transactions = client.instrumentation_store.get_all()
    assert transactions == []


@pytest.mark.skipif(travis_and_psycopg2,
                    reason="Requires postgres server. Only runs on travisci.")
def test_psycopg2_select_LIKE(postgres_connection):
    """
    Check that we pass queries with %-notation but without parameters
    properly to the dbapi backend
    """
    client = get_client()
    control.instrument()
    cursor = postgres_connection.cursor()

    try:
        client.begin_transaction("web.django")
        cursor.execute("SELECT * FROM test WHERE name LIKE 't%'")
        cursor.fetchall()
        client.end_transaction(None, "test-transaction")
    finally:
        # make sure we've cleared out the traces for the other tests.
        transactions = client.instrumentation_store.get_all()
        traces = transactions[0]['traces']
        assert traces[0]['name'] == 'SELECT FROM test'



"""
elasticapm.contrib.celery
~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
try:
    from celery.task import task
except ImportError:
    from celery.decorators import task
from celery import signals


class CeleryFilter(object):
    def filter(self, record):
        if record.funcName in ('_log_error',):
            return 0
        else:
            return 1


def register_signal(client):
    def process_failure_signal(sender, task_id, exception, args, kwargs,
                               traceback, einfo, **kw):
        client.capture_exception(
            extra={
                'task_id': task_id,
                'task': sender,
                'args': args,
                'kwargs': kwargs,
            })
    signals.task_failure.connect(process_failure_signal, weak=False)



"""
elasticapm.handlers.logbook
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
from __future__ import absolute_import

import sys
import traceback

import logbook

from elasticapm.base import Client
from elasticapm.utils import six
from elasticapm.utils.encoding import to_string

LOOKBOOK_LEVELS = {
    logbook.DEBUG: 'debug',
    logbook.INFO: 'info',
    logbook.NOTICE: 'info',
    logbook.WARNING: 'warning',
    logbook.ERROR: 'error',
    logbook.CRITICAL: 'fatal',
}


class LogbookHandler(logbook.Handler):
    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            arg = args[0]
            # if isinstance(arg, six.string_types):
            # self.client = kwargs.pop('client_cls', Client)(dsn=arg)
            if isinstance(arg, Client):
                self.client = arg
            else:
                raise ValueError(
                    'The first argument to %s must be a Client instance, '
                    'got %r instead.' % (
                        self.__class__.__name__,
                        arg,
                    ))
            args = []
        else:
            try:
                self.client = kwargs.pop('client')
            except KeyError:
                raise TypeError(
                    'Expected keyword argument for LoggingHandler: client')
        super(LogbookHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        self.format(record)

        # Avoid typical config issues by overriding loggers behavior
        if record.channel.startswith('elasticapm.errors'):
            six.print_(to_string(record.message), file=sys.stderr)
            return

        try:
            return self._emit(record)
        except Exception:
            six.print_(sys.stderr,
                       "Top level ElasticAPM exception caught - "
                       "failed creating log record",
                       file=sys.stderr)
            six.print_(sys.stderr, to_string(record.msg), file=sys.stderr)
            six.print_(sys.stderr, to_string(traceback.format_exc()),
                       file=sys.stderr)

            try:
                self.client.capture('Exception')
            except Exception:
                pass

    def _emit(self, record):
        data = {
            'level': LOOKBOOK_LEVELS[record.level],
            'logger': record.channel,
        }

        # If there's no exception being processed,
        # exc_info may be a 3-tuple of None
        # http://docs.python.org/library/sys.html#sys.exc_info
        if record.exc_info is True or (
                    record.exc_info and all(record.exc_info)):
            handler = self.client.get_handler('elasticapm.events.Exception')

            data.update(handler.capture(self.client, exc_info=record.exc_info))

        return self.client.capture('Message',
                                   param_message=
                                   {
                                       'message': record.msg,
                                       'params': record.args
                                   },
                                   data=data,
                                   extra=record.extra,
        )



"""
elasticapm.middleware
~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import sys

from elasticapm.utils import get_url_dict
from elasticapm.utils.wsgi import get_current_url, get_environ, get_headers


class ElasticAPM(object):
    """
    A WSGI middleware which will attempt to capture any
    uncaught exceptions and send them to ElasticAPM.

    >>> from elasticapm.base import Client
    >>> application = ElasticAPM(application, Client())
    """
    def __init__(self, application, client):
        self.application = application
        self.client = client

    def __call__(self, environ, start_response):
        try:
            for event in self.application(environ, start_response):
                yield event
        except Exception:
            exc_info = sys.exc_info()
            self.handle_exception(exc_info, environ)
            exc_info = None
            raise

    def handle_exception(self, exc_info, environ):
        event_id = self.client.capture(
            'Exception',
            exc_info=exc_info,
            data={
                'context': {
                    'request': {
                        'method': environ.get('REQUEST_METHOD'),
                        'url': get_url_dict(get_current_url(environ)),
                        'headers': dict(get_headers(environ)),
                        'env': dict(get_environ(environ)),
                    }
                }
            }
        )
        return event_id



from __future__ import absolute_import

import sys
from optparse import make_option

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.core.management.color import color_style
from django.utils import termcolors

from elasticapm.contrib.django.models import get_client_class, get_client_config

try:
    from django.core.management.base import OutputWrapper
except ImportError:
    OutputWrapper = None


blue = termcolors.make_style(opts=('bold',), fg='blue')
cyan = termcolors.make_style(opts=('bold',), fg='cyan')
green = termcolors.make_style(fg='green')
magenta = termcolors.make_style(opts=('bold',), fg='magenta')
red = termcolors.make_style(opts=('bold',), fg='red')
white = termcolors.make_style(opts=('bold',), fg='white')
yellow = termcolors.make_style(opts=('bold',), fg='yellow')


class TestException(Exception):
    pass


class ColoredLogger(object):
    def __init__(self, stream):
        self.stream = stream
        self.errors = []
        self.color = color_style()

    def log(self, level, *args, **kwargs):
        style = kwargs.pop('style', self.color.NOTICE)
        msg = ' '.join((level.upper(), args[0] % args[1:], '\n'))
        if OutputWrapper is None:
            self.stream.write(msg)
        else:
            self.stream.write(msg, style_func=style)

    def error(self, *args, **kwargs):
        kwargs['style'] = red
        self.log('error', *args, **kwargs)
        self.errors.append((args,))

    def warning(self, *args, **kwargs):
        kwargs['style'] = yellow
        self.log('warning', *args, **kwargs)

    def info(self, *args, **kwargs):
        kwargs['style'] = green
        self.log('info', *args, **kwargs)


LOGO = """

                              .o8                               .
                             "888                             .o8
         .ooooo.  oo.ooooo.   888oooo.   .ooooo.   .oooo.   .o888oo
        d88' `88b  888' `88b  d88' `88b d88' `88b `P  )88b    888
        888   888  888   888  888   888 888ooo888  .oP"888    888
        888   888  888   888  888   888 888    .o d8(  888    888 .
        `Y8bod8P'  888bod8P'  `Y8bod8P' `Y8bod8P' `Y888""8o   "888"
                   888
                  o888o

"""


CONFIG_EXAMPLE = """

You can set it in your settings file:

    ELASTICAPM = {
        'ORGANIZATION_ID': '<YOUR-ORGANIZATION-ID>',
        'APP_ID': '<YOUR-APP-ID>',
        'SECRET_TOKEN': '<YOUR-SECRET-TOKEN>',
    }

or with environment variables:

    $ export ELASTICAPM_APP_ID="<YOUR-APP-ID>"
    $ export ELASTICAPM_SECRET_TOKEN="<YOUR-SECRET-TOKEN>"
    $ python manage.py elasticapm check

"""


class Command(BaseCommand):
    arguments = (
        (('-a', '--app-name'),
         {'default': None, 'dest': 'app_name', 'help': 'Specifies the app name.'}
        ),

        (('-t', '--token'),
         {'default': None, 'dest': 'secret_token',
          'help': 'Specifies the secret token.'}
         )
    )
    if not hasattr(BaseCommand, 'add_arguments'):
        # Django <= 1.7
        option_list = BaseCommand.option_list + tuple(
            make_option(*args, **kwargs) for args, kwargs in arguments
        )
    args = 'test check'

    # Django 1.8+
    def add_arguments(self, parser):
        parser.add_argument('subcommand')
        for args, kwargs in self.arguments:
            parser.add_argument(*args, **kwargs)

    def handle(self, *args, **options):
        if 'subcommand' in options:
            # Django 1.8+, argparse
            subcommand = options['subcommand']
        elif not args:
            return self.handle_command_not_found('No command specified.')
        else:
            # Django 1.7, optparse
            subcommand = args[0]
        if subcommand not in self.dispatch:
            self.handle_command_not_found('No such command "%s".' % subcommand)
        else:
            self.dispatch.get(
                subcommand,
                self.handle_command_not_found
            )(self, subcommand, **options)

    def handle_test(self, command, **options):
        """Send a test error to APM Server"""
        self.write(LOGO, cyan)
        config = get_client_config()
        # can't be async for testing
        config['async_mode'] = False
        for key in ('app_name', 'secret_token'):
            if options.get(key):
                config[key] = options[key]
        client_class = get_client_class()
        client = client_class(**config)
        client.error_logger = ColoredLogger(self.stderr)
        client.logger = ColoredLogger(self.stderr)
        client.state.logger = client.logger
        client.state.error_logger = client.error_logger
        self.write(
            "Trying to send a test error to APM Server using these settings:\n\n"
            "APP_NAME:\t\t\t%s\n"
            "SECRET_TOKEN:\t\t%s\n"
            "SERVERS:\t\t%s\n\n" % (
                client.app_name,
                client.secret_token,
                ', '.join(client.servers)
            )
        )

        try:
            raise TestException('Hi there!')
        except TestException as e:
            result = client.capture_exception()
            if not client.error_logger.errors:
                self.write(
                    'Success! We tracked the error successfully! You should be'
                    ' able to see it in a few seconds at the above URL'
                )

    def handle_check(self, command, **options):
        """Check your settings for common misconfigurations"""
        self.write(LOGO, cyan)
        passed = True
        config = get_client_config()
        client_class = get_client_class()
        client = client_class(**config)
        # check if org/app and token are set:
        is_set = lambda x: x and x != 'None'
        values = [client.app_name, client.secret_token]
        if all(map(is_set, values)):
            self.write(
                'App name and secret token are set, good job!',
                green
            )
        else:
            passed = False
            self.write(
                'Configuration errors detected!', red, ending='\n\n'
            )
            if not is_set(client.app_name):
                self.write("  * APP_NAME not set! ", red, ending='\n')
            if not is_set(client.secret_token):
                self.write("  * SECRET_TOKEN not set!", red, ending='\n')
            self.write(CONFIG_EXAMPLE)
        self.write('')

        # check if we're disabled due to DEBUG:
        if settings.DEBUG:
            if getattr(settings, 'ELASTICAPM', {}).get('DEBUG'):
                self.write(
                    'Note: even though you are running in DEBUG mode, we will '
                    'send data to the APM Server, because you set ELASTICAPM["DEBUG"] to '
                    'True. You can disable ElasticAPM while in DEBUG mode like this'
                    '\n\n',
                    yellow
                )
                self.write(
                    '   ELASTICAPM = {\n'
                    '       "DEBUG": False,\n'
                    '       # your other ELASTICAPM settings\n'
                    '   }'

                )
            else:
                self.write(
                    'Looks like you\'re running in DEBUG mode. ElasticAPM will NOT '
                    'gather any data while DEBUG is set to True.\n\n',
                    red,
                )
                self.write(
                    'If you want to test ElasticAPM while DEBUG is set to True, you'
                    ' can force ElasticAPM to gather data by setting'
                    ' ELASTICAPM["DEBUG"] to True, like this\n\n'
                    '   ELASTICAPM = {\n'
                    '       "DEBUG": True,\n'
                    '       # your other ELASTICAPM settings\n'
                    '   }'
                )
                passed = False
        else:
            self.write(
                'DEBUG mode is disabled! Looking good!',
                green
            )
        self.write('')

        # check if middleware is set, and if it is at the first position
        middleware = list(settings.MIDDLEWARE_CLASSES)
        try:
            pos = middleware.index(
                'elasticapm.contrib.django.middleware.TracingMiddleware'
            )
            if pos == 0:
                self.write(
                    'Tracing middleware is configured! Awesome!',
                    green
                )
            else:
                self.write(
                    'Tracing middleware is configured, but not at the first '
                    'position\n',
                    yellow
                )
                self.write(
                    'ElasticAPM works best if you add it at the top of your '
                    'MIDDLEWARE_CLASSES'
                )
        except ValueError:
            self.write(
                'Tracing middleware not configured!', red
            )
            self.write(
                '\n'
                'Add it to your MIDDLEWARE_CLASSES like this:\n\n'
                '    MIDDLEWARE_CLASSES = (\n'
                '        "elasticapm.contrib.django.middleware.TracingMiddleware",\n'
                '        # your other middleware classes\n'
                '    )\n'
            )
        self.write('')
        if passed:
            self.write('Looks like everything should be ready!', green)
        else:
            self.write(
                'Please fix the above errors. If you have any questions, write '
                'us at support@opbeat.com!',
                red
            )
        self.write('')
        return passed

    def handle_command_not_found(self, message):
        self.write(LOGO, cyan)
        self.write(message, red, ending='')
        self.write(
            ' Please use one of the following commands:\n\n',
            red
        )
        self.write(
            ''.join(
                ' * %s\t%s\n' % (k.ljust(8), v.__doc__)
                for k, v in self.dispatch.items()
            )
        )
        self.write('\n')
        argv = self._get_argv()
        self.write(
            'Usage:\n\t%s elasticapm <command>' % (
                ' '.join(argv[:argv.index('elasticapm')])
            )
        )

    def write(self, msg, style_func=None, ending=None, stream=None):
        """
        wrapper around self.stdout/stderr to ensure Django 1.4 compatibility
        """
        if stream is None:
            stream = self.stdout
        if OutputWrapper is None:
            ending = '\n' if ending is None else ending
            msg += ending
            stream.write(msg)
        else:
            stream.write(msg, style_func=style_func, ending=ending)

    def _get_argv(self):
        """allow cleaner mocking of sys.argv"""
        return sys.argv

    dispatch = {
        'test': handle_test,
        'check': handle_check,
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from elasticapm import processors
from elasticapm.utils import six
from tests.utils.compat import TestCase


class SanitizePasswordsProcessorTest(TestCase):
    def setUp(self):
        self.http_test_data = {
            'context': {
                'request': {
                    'body': 'foo=bar&password=123456&the_secret=abc&cc=1234567890098765',
                    'env': {
                        'foo': 'bar',
                        'password': 'hello',
                        'the_secret': 'hello',
                        'a_password_here': 'hello',
                    },
                    'headers': {
                        'foo': 'bar',
                        'password': 'hello',
                        'the_secret': 'hello',
                        'a_password_here': 'hello',
                    },
                    'cookies': {
                        'foo': 'bar',
                        'password': 'topsecret',
                        'the_secret': 'topsecret',
                        'sessionid': '123',
                        'a_password_here': '123456',
                    },
                    'url': {
                        'raw': 'http://example.com/bla?foo=bar&password=123456&the_secret=abc&cc=1234567890098765',
                        'search': 'foo=bar&password=123456&the_secret=abc&cc=1234567890098765'
                    }
                }
            }
        }

    def test_stacktrace(self):
        data = {
            'exception': {
                'stacktrace': [
                    {
                        'vars': {
                            'foo': 'bar',
                            'password': 'hello',
                            'the_secret': 'hello',
                            'a_password_here': 'hello',
                        },
                    }
                ]
            }
        }

        result = processors.sanitize_stacktrace_locals(None, data)

        self.assertTrue('stacktrace' in result['exception'])
        stack = result['exception']['stacktrace']
        self.assertEquals(len(stack), 1)
        frame = stack[0]
        self.assertTrue('vars' in frame)
        vars = frame['vars']
        self.assertTrue('foo' in vars)
        self.assertEquals(vars['foo'], 'bar')
        self.assertTrue('password' in vars)
        self.assertEquals(vars['password'], processors.MASK)
        self.assertTrue('the_secret' in vars)
        self.assertEquals(vars['the_secret'], processors.MASK)
        self.assertTrue('a_password_here' in vars)
        self.assertEquals(vars['a_password_here'], processors.MASK)

    def test_remove_http_request_body(self):
        assert 'body' in self.http_test_data['context']['request']

        result = processors.remove_http_request_body(None, self.http_test_data)

        assert 'body' not in result['context']['request']

    def test_sanitize_http_request_cookies(self):
        self.http_test_data['context']['request']['headers']['cookie'] =\
            'foo=bar; password=12345; the_secret=12345; csrftoken=abc'

        result = processors.sanitize_http_request_cookies(None, self.http_test_data)

        assert result['context']['request']['cookies'] == {
            'foo': 'bar',
            'password': processors.MASK,
            'the_secret': processors.MASK,
            'sessionid': processors.MASK,
            'a_password_here': processors.MASK,
        }

        assert (result['context']['request']['headers']['cookie'] ==
                'foo=bar; password={0}; the_secret={0}; csrftoken={0}'.format(processors.MASK))

    def test_sanitize_http_request_headers(self):
        result = processors.sanitize_http_request_headers(None, self.http_test_data)

        assert result['context']['request']['headers'] == {
            'foo': 'bar',
            'password': processors.MASK,
            'the_secret': processors.MASK,
            'a_password_here': processors.MASK,
        }

    def test_sanitize_http_wgi_env(self):
        result = processors.sanitize_http_wsgi_env(None, self.http_test_data)

        assert result['context']['request']['env'] == {
            'foo': 'bar',
            'password': processors.MASK,
            'the_secret': processors.MASK,
            'a_password_here': processors.MASK,
        }

    def test_sanitize_http_query_string(self):
        result = processors.sanitize_http_request_querystring(None, self.http_test_data)

        expected = 'foo=bar&password={0}&the_secret={0}&cc={0}'.format(
            processors.MASK
        )
        assert result['context']['request']['url']['search'] == expected
        assert result['context']['request']['url']['raw'].endswith(expected)

    def test_post_as_string(self):
        result = processors.sanitize_http_request_body(None, self.http_test_data)
        expected = 'foo=bar&password={0}&the_secret={0}&cc={0}'.format(
            processors.MASK
        )
        assert result['context']['request']['body'] == expected

    def test_querystring_as_string_with_partials(self):
        self.http_test_data['context']['request']['url']['search'] = 'foo=bar&password&secret=123456'
        result = processors.sanitize_http_request_querystring(None, self.http_test_data)

        assert (result['context']['request']['url']['search'] ==
                'foo=bar&password&secret={0}'.format(processors.MASK))

    def test_sanitize_credit_card(self):
        result = processors._sanitize('foo', '4242424242424242')
        self.assertEquals(result, processors.MASK)

    def test_sanitize_credit_card_with_spaces(self):
        result = processors._sanitize('foo', '4242 4242 4242 4242')
        self.assertEquals(result, processors.MASK)

    def test_non_utf8_encoding(self):
        broken = six.b('broken=') + u"aéöüa".encode('latin-1')
        self.http_test_data['context']['request']['url']['search'] = broken
        result = processors.sanitize_http_request_querystring(None, self.http_test_data)
        assert result['context']['request']['url']['search'] == u'broken=a\ufffd\ufffd\ufffda'


def test_remove_http_request_body():
    data = {
        'context': {
            'request': {
                'body': 'foo'
            },
        }
    }

    result = processors.remove_http_request_body(None, data)

    assert 'request' in result['context']
    request = result['context']['request']
    assert 'body' not in request


def test_remove_stacktrace_locals():
    data = {
        'exception': {
            'stacktrace': [
                {
                    'vars': {
                        'foo': 'bar',
                        'password': 'hello',
                        'the_secret': 'hello',
                        'a_password_here': 'hello',
                    },
                }
            ]
        }
    }
    result = processors.remove_stacktrace_locals(None, data)

    assert 'stacktrace' in result['exception']
    stack = result['exception']['stacktrace']
    for frame in stack:
        assert 'vars' not in frame



import memcache
import mock

import elasticapm
import elasticapm.instrumentation.control
from elasticapm.traces import trace
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class InstrumentMemcachedTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        elasticapm.instrumentation.control.instrument()

    @mock.patch("elasticapm.traces.TransactionsStore.should_collect")
    def test_memcached(self, should_collect):
        should_collect.return_value = False
        self.client.begin_transaction("transaction.test")
        with trace("test_memcached", "test"):
            conn = memcache.Client(['127.0.0.1:11211'], debug=0)
            conn.set("mykey", "a")
            assert "a" == conn.get("mykey")
            assert {"mykey": "a"} == conn.get_multi(["mykey", "myotherkey"])
        self.client.end_transaction("BillingView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']

        expected_signatures = ['transaction', 'test_memcached',
                               'Client.set', 'Client.get',
                               'Client.get_multi']

        self.assertEqual(set([t['name'] for t in traces]),
                         set(expected_signatures))

        self.assertEqual(traces[0]['name'], 'Client.set')
        self.assertEqual(traces[0]['type'], 'cache.memcached')
        self.assertEqual(traces[0]['parent'], 1)

        self.assertEqual(traces[1]['name'], 'Client.get')
        self.assertEqual(traces[1]['type'], 'cache.memcached')
        self.assertEqual(traces[1]['parent'], 1)

        self.assertEqual(traces[2]['name'], 'Client.get_multi')
        self.assertEqual(traces[2]['type'], 'cache.memcached')
        self.assertEqual(traces[2]['parent'], 1)

        self.assertEqual(traces[3]['name'], 'test_memcached')
        self.assertEqual(traces[3]['type'], 'test')

        self.assertEqual(traces[4]['name'], 'transaction')
        self.assertEqual(traces[4]['type'], 'transaction')

        self.assertEqual(len(traces), 5)



from elasticapm.instrumentation.packages.dbapi2 import (ConnectionProxy,
                                                        CursorProxy,
                                                        DbApi2Instrumentation,
                                                        extract_signature)
from elasticapm.traces import trace
from elasticapm.utils import default_ports


class PGCursorProxy(CursorProxy):
    provider_name = 'postgresql'

    def extract_signature(self, sql):
        return extract_signature(sql)


class PGConnectionProxy(ConnectionProxy):
    cursor_proxy = PGCursorProxy


class Psycopg2Instrumentation(DbApi2Instrumentation):
    name = 'psycopg2'

    instrument_list = [
        ("psycopg2", "connect")
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        signature = "psycopg2.connect"

        host = kwargs.get('host')
        if host:
            signature += " " + str(host)

            port = kwargs.get('port')
            if port:
                port = str(port)
                if int(port) != default_ports.get("postgresql"):
                    signature += ":" + port
        else:
            # Parse connection string and extract host/port
            pass

        with trace(signature, "db.postgreql.connect"):
            return PGConnectionProxy(wrapped(*args, **kwargs))


class Psycopg2RegisterTypeInstrumentation(DbApi2Instrumentation):
    name = 'psycopg2-register-type'

    instrument_list = [
        ("psycopg2.extensions", "register_type"),
        # specifically instrument `register_json` as it bypasses `register_type`
        ("psycopg2._json", "register_json"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        if ('conn_or_curs' in kwargs and
                hasattr(kwargs['conn_or_curs'], "__wrapped__")):
            kwargs['conn_or_curs'] = kwargs['conn_or_curs'].__wrapped__
        # register_type takes the connection as second argument
        elif len(args) == 2 and hasattr(args[1], "__wrapped__"):
                args = (args[0], args[1].__wrapped__)
        # register_json takes the connection as first argument, and can have
        # several more arguments
        elif method == 'register_json':
            if args and hasattr(args[0], "__wrapped__"):
                args = (args[0].__wrapped__,) + args[1:]

        return wrapped(*args, **kwargs)



"""
elasticapm.contrib.django.middleware.wsgi
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from elasticapm.middleware import ElasticAPM as ElasticAPMBase


class ElasticAPM(ElasticAPMBase):
    """
    Identical to the default WSGI middleware except that
    the client comes dynamically via ``get_client

    >>> from elasticapm.contrib.django.middleware.wsgi import ElasticAPM
    >>> application = ElasticAPM(application)
    """
    def __init__(self, application):
        self.application = application

    @property
    def client(self):
        from elasticapm.contrib.django.models import client
        return client



from elasticapm.instrumentation.packages.dbapi2 import (ConnectionProxy,
                                                        CursorProxy,
                                                        DbApi2Instrumentation,
                                                        extract_signature)
from elasticapm.traces import trace


class SQLiteCursorProxy(CursorProxy):
    provider_name = 'sqlite'

    def extract_signature(self, sql):
        return extract_signature(sql)


class SQLiteConnectionProxy(ConnectionProxy):
    cursor_proxy = SQLiteCursorProxy


class SQLiteInstrumentation(DbApi2Instrumentation):
    name = 'sqlite'

    instrument_list = [
        ("sqlite3", "connect"),
        ("sqlite3.dbapi2", "connect"),
        ("pysqlite2.dbapi2", "connect"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        signature = ".".join([module, method])

        if len(args) == 1:
            signature += " " + str(args[0])

        with trace(signature, "db.sqlite.connect"):
            return SQLiteConnectionProxy(wrapped(*args, **kwargs))



"""
elasticapm.contrib.async_worker
~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
import logging
import os
import time
from threading import Lock, Thread

from elasticapm.utils import six
from elasticapm.utils.six.moves import queue

logger = logging.getLogger('elasticapm')

ELASTICAPM_WAIT_SECONDS = 10


class AsyncWorker(object):
    _terminator = object()

    def __init__(self):
        self._queue = queue.Queue(-1)
        self._lock = Lock()
        self._thread = None
        self.start()

    def main_thread_terminated(self):
        self._lock.acquire()
        try:
            if not self._thread:
                # thread not started or already stopped - nothing to do
                return

            # wake the processing thread up
            self._queue.put_nowait(self._terminator)

            # wait briefly, initially
            initial_timeout = 0.1

            if not self._timed_queue_join(initial_timeout):
                # if that didn't work, wait a bit longer
                # NB that size is an approximation, because other threads may
                # add or remove items
                size = self._queue.qsize()

                six.print_(
                    "PID %i: ElasticAPM is attempting to send %i pending messages" % (
                        os.getpid(),
                        size,
                    )
                )
                six.print_(
                    "Waiting up to %s seconds, " % ELASTICAPM_WAIT_SECONDS,
                    end=''
                )
                wait_start = time.time()
                if os.name == 'nt':
                    six.print_("press Ctrl-Break to quit")
                else:
                    six.print_("press Ctrl-C to quit")
                self._timed_queue_join(ELASTICAPM_WAIT_SECONDS - initial_timeout)
                six.print_('PID %i: done, took %.2f seconds to complete' % (
                    os.getpid(),
                    time.time() - wait_start,
                ))
            self._thread = None

        finally:
            self._lock.release()

    def _timed_queue_join(self, timeout):
        """
        implementation of Queue.join which takes a 'timeout' argument

        returns True on success, False on timeout
        """
        deadline = time.time() + timeout

        with self._queue.all_tasks_done:
            while self._queue.unfinished_tasks:
                delay = deadline - time.time()
                if delay <= 0:
                    # timed out
                    return False

                self._queue.all_tasks_done.wait(timeout=delay)

            return True

    def start(self):
        """
        Starts the task thread.
        """
        self._lock.acquire()
        try:
            if not self._thread:
                self._thread = Thread(target=self._target)
                self._thread.setDaemon(True)
                self._thread.start()
        finally:
            self._lock.release()

    def stop(self, timeout=None):
        """
        Stops the task thread. Synchronous!
        """
        self._lock.acquire()
        try:
            if self._thread:
                self._queue.put_nowait(self._terminator)
                self._thread.join(timeout=timeout)
                self._thread = None
        finally:
            self._lock.release()

    def is_alive(self):
        return self._thread is not None and self._thread.is_alive()

    def queue(self, callback, kwargs):
        self._queue.put_nowait((callback, kwargs))

    def _target(self):
        while True:
            try:
                record = self._queue.get()
                if record is self._terminator:
                    break
                callback, kwargs = record
                try:
                    callback(**kwargs)
                except Exception:
                    logger.error('Error while sending', exc_info=True)
            finally:
                self._queue.task_done()
            time.sleep(0)


class Worker(object):
    """
    A WSGI middleware which provides ``environ['elasticapm.worker']``
    that can be used by clients to process asynchronous tasks.

    >>> from elasticapm.base import Client
    >>> application = Worker(application)
    """
    def __init__(self, application):
        self.application = application
        self.worker = AsyncWorker()

    def __call__(self, environ, start_response):
        environ['elasticapm.worker'] = self.worker
        for event in self.application(environ, start_response):
            yield event



"""
This module implements WSGI related helpers adapted from ``werkzeug.wsgi``

:copyright: (c) 2010 by the Werkzeug Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
from elasticapm.utils import six

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote


# `get_headers` comes from `werkzeug.datastructures.EnvironHeaders`
def get_headers(environ):
    """
    Returns only proper HTTP headers.
    """
    for key, value in six.iteritems(environ):
        key = str(key)
        if key.startswith('HTTP_') and key not in \
           ('HTTP_CONTENT_TYPE', 'HTTP_CONTENT_LENGTH'):
            yield key[5:].replace('_', '-').lower(), value
        elif key in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
            yield key.replace('_', '-').lower(), value


def get_environ(environ):
    """
    Returns our whitelisted environment variables.
    """
    for key in ('REMOTE_ADDR', 'SERVER_NAME', 'SERVER_PORT'):
        if key in environ:
            yield key, environ[key]


# `get_host` comes from `werkzeug.wsgi`
def get_host(environ):
    """Return the real host for the given WSGI environment.  This takes care
    of the `X-Forwarded-Host` header.

    :param environ: the WSGI environment to get the host of.
    """
    scheme = environ.get('wsgi.url_scheme')
    if 'HTTP_X_FORWARDED_HOST' in environ:
        result = environ['HTTP_X_FORWARDED_HOST']
    elif 'HTTP_HOST' in environ:
        result = environ['HTTP_HOST']
    else:
        result = environ['SERVER_NAME']
        if (scheme, str(environ['SERVER_PORT'])) not \
           in (('https', '443'), ('http', '80')):
            result += ':' + environ['SERVER_PORT']
    if result.endswith(':80') and scheme == 'http':
        result = result[:-3]
    elif result.endswith(':443') and scheme == 'https':
        result = result[:-4]
    return result


# `get_current_url` comes from `werkzeug.wsgi`
def get_current_url(environ, root_only=False, strip_querystring=False,
                    host_only=False):
    """A handy helper function that recreates the full URL for the current
    request or parts of it.  Here an example:

    >>> from werkzeug import create_environ
    >>> env = create_environ("/?param=foo", "http://localhost/script")
    >>> get_current_url(env)
    'http://localhost/script/?param=foo'
    >>> get_current_url(env, root_only=True)
    'http://localhost/script/'
    >>> get_current_url(env, host_only=True)
    'http://localhost/'
    >>> get_current_url(env, strip_querystring=True)
    'http://localhost/script/'

    :param environ: the WSGI environment to get the current URL from.
    :param root_only: set `True` if you only want the root URL.
    :param strip_querystring: set to `True` if you don't want the querystring.
    :param host_only: set to `True` if the host URL should be returned.
    """
    tmp = [environ['wsgi.url_scheme'], '://', get_host(environ)]
    cat = tmp.append
    if host_only:
        return ''.join(tmp) + '/'
    cat(quote(environ.get('SCRIPT_NAME', '').rstrip('/')))
    if root_only:
        cat('/')
    else:
        cat(quote('/' + environ.get('PATH_INFO', '').lstrip('/')))
        if not strip_querystring:
            qs = environ.get('QUERY_STRING')
            if qs:
                cat('?' + qs)
    return ''.join(tmp)



from twisted.logger import ILogObserver
from zope.interface import implementer

from elasticapm.base import Client


@implementer(ILogObserver)
class LogObserver(object):
    """
    A twisted log observer for Elastic APM.
    Eg.:

    from elasticapm.base import Client
    from twisted.logger import Logger

    client = Client(...)
    observer = LogObserver(client=client)
    log = Logger(observer=observer)

    try:
        1 / 0
    except:
        log.failure("Math is hard!")
    """

    def __init__(self, client=None, **kwargs):
        self.client = client or Client(**kwargs)

    def __call__(self, event):
        failure = event.get('log_failure')
        if failure is not None:
            self.client.capture_exception(
                (failure.type, failure.value, failure.getTracebackObject()),
                extra=event)



import socket

import mock
import pytest

from elasticapm.transport.base import TransportException
from elasticapm.transport.http import HTTPTransport
from elasticapm.utils import six
from elasticapm.utils.compat import HTTPError, urlparse
from tests.utils.compat import TestCase


class TestHttpFailures(TestCase):
    @mock.patch('elasticapm.transport.http.urlopen')
    def test_send(self, mock_urlopen):
        transport = HTTPTransport(urlparse.urlparse('http://localhost:9999'))
        mock_response = mock.Mock(
            info=lambda: {'Location': 'http://example.com/foo'}
        )
        mock_urlopen.return_value = mock_response
        url = transport.send('x', {})
        assert url == 'http://example.com/foo'
        assert mock_response.close.call_count == 1

    @mock.patch('elasticapm.transport.http.urlopen')
    def test_timeout(self, mock_urlopen):
        transport = HTTPTransport(urlparse.urlparse('http://localhost:9999'))
        mock_urlopen.side_effect = socket.timeout()
        with pytest.raises(TransportException) as exc_info:
            transport.send('x', {})
        assert 'timeout' in str(exc_info.value)

    @mock.patch('elasticapm.transport.http.urlopen')
    def test_http_error(self, mock_urlopen):
        url, status, message, body = (
            'http://localhost:9999', 418, "I'm a teapot", 'Nothing'
        )
        transport = HTTPTransport(urlparse.urlparse(url))
        mock_urlopen.side_effect = HTTPError(
            url, status, message, hdrs={}, fp=six.StringIO(body)
        )
        with pytest.raises(TransportException) as exc_info:
            transport.send('x', {})
        for val in (url, status, message, body):
            assert str(val) in str(exc_info.value)

    @mock.patch('elasticapm.transport.http.urlopen')
    def test_generic_error(self, mock_urlopen):
        url, status, message, body = (
            'http://localhost:9999', 418, "I'm a teapot", 'Nothing'
        )
        transport = HTTPTransport(urlparse.urlparse(url))
        mock_urlopen.side_effect = Exception('Oopsie')
        with pytest.raises(TransportException) as exc_info:
            transport.send('x', {})
        assert 'Oopsie' in str(exc_info.value)



import mock
import requests
from requests.exceptions import InvalidURL, MissingSchema

import elasticapm
import elasticapm.instrumentation.control
from elasticapm.traces import trace
from tests.helpers import get_tempstoreclient
from tests.utils.compat import TestCase


class InstrumentRequestsTest(TestCase):
    def setUp(self):
        self.client = get_tempstoreclient()
        elasticapm.instrumentation.control.instrument()

    @mock.patch("requests.adapters.HTTPAdapter.send")
    def test_requests_instrumentation(self, mock_send):
        mock_send.return_value = mock.Mock(
            url='http://example.com',
            history=[],
            headers={'location': ''},
        )
        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            # NOTE: The `allow_redirects` argument has to be set to `False`,
            # because mocking is done a level deeper, and the mocked response
            # from the `HTTPAdapter` is about to be used to make further
            # requests to resolve redirects, which doesn't make sense for this
            # test case.
            requests.get('http://example.com', allow_redirects=False)
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']
        self.assertEqual('GET example.com', traces[0]['name'])
        self.assertEqual('http://example.com/', traces[0]['context']['url'])

    @mock.patch("requests.adapters.HTTPAdapter.send")
    def test_requests_instrumentation_via_session(self, mock_send):
        mock_send.return_value = mock.Mock(
            url='http://example.com',
            history=[],
            headers={'location': ''},
        )
        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            s = requests.Session()
            s.get('http://example.com', allow_redirects=False)
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']
        self.assertEqual('GET example.com', traces[0]['name'])
        self.assertEqual('http://example.com/', traces[0]['context']['url'])

    @mock.patch("requests.adapters.HTTPAdapter.send")
    def test_requests_instrumentation_via_prepared_request(self, mock_send):
        mock_send.return_value = mock.Mock(
            url='http://example.com',
            history=[],
            headers={'location': ''},
        )
        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            r = requests.Request('get', 'http://example.com')
            pr = r.prepare()
            s = requests.Session()
            s.send(pr, allow_redirects=False)
        self.client.end_transaction("MyView")

        transactions = self.client.instrumentation_store.get_all()
        traces = transactions[0]['traces']
        self.assertEqual('GET example.com', traces[0]['name'])
        self.assertEqual('http://example.com/', traces[0]['context']['url'])

    def test_requests_instrumentation_malformed_none(self):
        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            self.assertRaises(MissingSchema, requests.get, None)

    def test_requests_instrumentation_malformed_schema(self):
        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            self.assertRaises(MissingSchema, requests.get, '')

    def test_requests_instrumentation_malformed_path(self):
        self.client.begin_transaction("transaction.test")
        with trace("test_pipeline", "test"):
            self.assertRaises(InvalidURL, requests.get, 'http://')



from __future__ import absolute_import

from elasticapm.contrib.pylons import ElasticAPM
from tests.utils.compat import TestCase


def example_app(environ, start_response):
    raise ValueError('hello world')


class MiddlewareTest(TestCase):
    def setUp(self):
        self.app = example_app

    def test_init(self):
        config = {
            'elasticapm.servers': 'http://localhost/api/store',
            'elasticapm.app_name': 'p' * 32,
            'elasticapm.secret_token': 'a' * 32,
        }
        middleware = ElasticAPM(self.app, config)



"""
elasticapm.base
~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import absolute_import

import datetime
import logging
import os
import platform
import socket
import sys
import time
import uuid
import warnings
import zlib

import elasticapm
from elasticapm.conf import defaults
from elasticapm.traces import TransactionsStore, get_transaction
from elasticapm.transport.base import TransportException
from elasticapm.utils import json_encoder as json
from elasticapm.utils import is_master_process, six, stacks, varmap
from elasticapm.utils.compat import atexit_register, urlparse
from elasticapm.utils.deprecation import deprecated
from elasticapm.utils.encoding import shorten, transform
from elasticapm.utils.module_import import import_string
from elasticapm.utils.stacks import get_culprit, iter_stack_frames

__all__ = ('Client',)


class ClientState(object):
    ONLINE = 1
    ERROR = 0

    def __init__(self):
        self.status = self.ONLINE
        self.last_check = None
        self.retry_number = 0

    def should_try(self):
        if self.status == self.ONLINE:
            return True

        interval = min(self.retry_number, 6) ** 2

        if time.time() - self.last_check > interval:
            return True

        return False

    def set_fail(self):
        self.status = self.ERROR
        self.retry_number += 1
        self.last_check = time.time()

    def set_success(self):
        self.status = self.ONLINE
        self.last_check = None
        self.retry_number = 0

    def did_fail(self):
        return self.status == self.ERROR


class Client(object):
    """
    The base ElasticAPM client, which handles communication over the
    HTTP API to the APM Server.

    Will read default configuration from the environment variable
    ``ELASTICAPM_APP_NAME`` and ``ELASTICAPM_SECRET_TOKEN``
    if available. ::

    >>> from elasticapm import Client

    >>> # Read configuration from environment
    >>> client = Client()

    >>> # Configure the client manually
    >>> client = Client(
    >>>     include_paths=['my.package'],
    >>>     app_name='app_name',
    >>>     secret_token='secret_token',
    >>> )

    >>> # Record an exception
    >>> try:
    >>>     1/0
    >>> except ZeroDivisionError:
    >>>     ident = client.get_ident(client.capture_exception())
    >>>     print ("Exception caught; reference is %%s" %% ident)
    """
    logger = logging.getLogger('elasticapm')
    protocol_version = '1.0'

    environment_config_map = {
        'app_name': 'ELASTICAPM_APP_NAME',
        'secret_token': 'ELASTICAPM_SECRET_TOKEN',
        'git_ref': 'ELASTICAPM_GIT_REF',
        'app_version': 'ELASTICAPM_APP_VERSION',
    }

    def __init__(self, app_name=None, secret_token=None,
                 transport_class=None, include_paths=None, exclude_paths=None,
                 timeout=None, hostname=None, auto_log_stacks=None,
                 string_max_length=None, list_max_length=None, processors=None,
                 filter_exception_types=None, servers=None,
                 async_mode=None, traces_send_freq_secs=None,
                 transactions_ignore_patterns=None, git_ref=None,
                 app_version=None,
                 **kwargs):
        self.app_name = self.secret_token = self.git_ref = self.app_version = None
        # configure loggers first
        cls = self.__class__
        self.logger = logging.getLogger('%s.%s' % (cls.__module__, cls.__name__))
        self.error_logger = logging.getLogger('elasticapm.errors')
        self.state = ClientState()
        self._configure(app_name=app_name,
                        secret_token=secret_token, git_ref=git_ref,
                        app_version=app_version)
        self.servers = servers or defaults.SERVERS
        self.async_mode = (async_mode is True
                           or (defaults.ASYNC_MODE and async_mode is not False))
        if not transport_class:
            transport_class = (defaults.ASYNC_TRANSPORT_CLASS
                               if self.async_mode
                               else defaults.SYNC_TRANSPORT_CLASS)
        self._transport_class = import_string(transport_class)
        self._transports = {}

        # servers may be set to a NoneType (for Django)
        if self.servers and not (self.app_name and self.secret_token):
            msg = 'Missing configuration for ElasticAPM client. Please see documentation.'
            self.logger.info(msg)

        self.is_send_disabled = (
            os.environ.get('ELASTICAPM_DISABLE_SEND', '').lower() in ('1', 'true')
        )
        if self.is_send_disabled:
            self.logger.info(
                'Not sending any data to APM Server due to ELASTICAPM_DISABLE_SEND '
                'environment variable'
            )

        self.include_paths = set(include_paths or defaults.INCLUDE_PATHS)
        self.exclude_paths = set(exclude_paths or defaults.EXCLUDE_PATHS)
        self.timeout = int(timeout or defaults.TIMEOUT)
        self.hostname = six.text_type(hostname or defaults.HOSTNAME)
        self.auto_log_stacks = bool(auto_log_stacks or
                                    defaults.AUTO_LOG_STACKS)

        self.string_max_length = int(string_max_length or
                                     defaults.MAX_LENGTH_STRING)
        self.list_max_length = int(list_max_length or defaults.MAX_LENGTH_LIST)
        self.traces_send_freq_secs = (traces_send_freq_secs or
                                      defaults.TRACES_SEND_FREQ_SECS)

        self.filter_exception_types_dict = {}
        for exc_to_filter in (filter_exception_types or []):
            exc_to_filter_type = exc_to_filter.split(".")[-1]
            exc_to_filter_module = ".".join(exc_to_filter.split(".")[:-1])
            self.filter_exception_types_dict[exc_to_filter_type] = exc_to_filter_module

        if processors is None:
            self.processors = defaults.PROCESSORS
        else:
            self.processors = processors
        self.processors = [import_string(p) for p in self.processors]

        self.instrumentation_store = TransactionsStore(
            lambda: self.get_stack_info_for_trace(iter_stack_frames(), False),
            self.traces_send_freq_secs,
            transactions_ignore_patterns
        )
        atexit_register(self.close)

    def _configure(self, **kwargs):
        """
        Configures this instance based on kwargs, or environment variables.
        The former take precedence.
        """
        for attr_name, value in kwargs.items():
            if value is None and (attr_name in self.environment_config_map
                                  and self.environment_config_map[attr_name] in os.environ):
                self.logger.info("Configuring elasticapm.%s from environment variable '%s'",
                                 attr_name, self.environment_config_map[attr_name])
                value = os.environ[self.environment_config_map[attr_name]]
            setattr(self, attr_name, six.text_type(value))

    def get_ident(self, result):
        """
        Returns a searchable string representing a message.

        >>> result = client.process(**kwargs)
        >>> ident = client.get_ident(result)
        """
        return result

    def get_handler(self, name):
        return import_string(name)

    def get_stack_info_for_trace(self, frames, extended=True):
        """Overrideable in derived clients to add frames/info, e.g. templates

        4.0: Use for error frames too.
        """
        return stacks.get_stack_info(frames, extended)

    def build_msg_for_logging(self, event_type, data=None, date=None,
                              extra=None, stack=None,
                              **kwargs):
        """
        Captures, processes and serializes an event into a dict object
        """
        # create ID client-side so that it can be passed to application
        event_id = uuid.uuid4().hex

        if data is None:
            data = {}
        if extra is None:
            extra = {}
        if not date:
            date = datetime.datetime.utcnow()
        if stack is None:
            stack = self.auto_log_stacks
        if 'context' not in data:
            data['context'] = context = {}
        else:
            context = data['context']

        # if '.' not in event_type:
        # Assume it's a builtin
        event_type = 'elasticapm.events.%s' % event_type

        handler = self.get_handler(event_type)
        result = handler.capture(self, data=data, **kwargs)
        if self._filter_exception_type(result):
            return
        # data (explicit) culprit takes over auto event detection
        culprit = result.pop('culprit', None)
        if data.get('culprit'):
            culprit = data['culprit']

        for k, v in six.iteritems(result):
            if k not in data:
                data[k] = v

        log = data.get('log', {})
        if stack and 'stacktrace' not in log:
            if stack is True:
                frames = iter_stack_frames()
            else:
                frames = stack
            frames = varmap(lambda k, v: shorten(
                v,
                string_length=self.string_max_length,
                list_length=self.list_max_length
            ), stacks.get_stack_info(frames))
            log['stacktrace'] = frames

        if 'stacktrace' in log and not culprit:
            culprit = get_culprit(
                log['stacktrace'],
                self.include_paths, self.exclude_paths
            )

        if 'level' in log and isinstance(log['level'], six.integer_types):
            log['level'] = logging.getLevelName(log['level']).lower()

        if log:
            data['log'] = log

        if culprit:
            data['culprit'] = culprit

        context['custom'] = extra

        # Run the data through processors
        for processor in self.processors:
            data = processor(self, data)

        # Make sure all data is coerced
        data = transform(data)

        data.update({
            'timestamp':  date.strftime(defaults.TIMESTAMP_FORMAT),
        })

        return self.build_msg({'errors': [data]})

    def build_msg(self, data=None, **kwargs):
        data = data or {}
        data['app'] = self.get_app_info()
        data['system'] = self.get_system_info()
        data.update(**kwargs)
        return data

    def capture(self, event_type, data=None, date=None,
                extra=None, stack=None, **kwargs):
        """
        Captures and processes an event and pipes it off to Client.send.

        To use structured data (interfaces) with capture:

        >>> client.capture('Message', message='foo', data={
        >>>     'http': {
        >>>         'url': '...',
        >>>         'data': {},
        >>>         'query_string': '...',
        >>>         'method': 'POST',
        >>>     },
        >>>     'logger': 'logger.name',
        >>>     'site': 'site.name',
        >>> }, extra={
        >>>     'key': 'value',
        >>> })

        The finalized ``data`` structure contains the following (some optional)
        builtin values:

        >>> {
        >>>     # the culprit and version information
        >>>     'culprit': 'full.module.name', # or /arbitrary/path
        >>>
        >>>     # arbitrary data provided by user
        >>>     'extra': {
        >>>         'key': 'value',
        >>>     }
        >>> }

        :param event_type: the module path to the Event class. Builtins can use
                           shorthand class notation and exclude the full module
                           path.
        :param data: the data base
        :param date: the datetime of this event
        :param extra: a dictionary of additional standard metadata
        :return: a 32-length string identifying this event
        """

        data = self.build_msg_for_logging(event_type, data, date,
                                          extra, stack, **kwargs)

        if data:
            servers = [server + defaults.ERROR_API_PATH for server in self.servers]
            self.send(servers=servers, **data)
            return data['errors'][0]['id']

    def _send_remote(self, url, data, headers=None):
        if headers is None:
            headers = {}
        parsed = urlparse.urlparse(url)
        transport = self._get_transport(parsed)
        if transport.async_mode:
            transport.send_async(
                data, headers,
                success_callback=self.handle_transport_success,
                fail_callback=self.handle_transport_fail
            )
        else:
            url = transport.send(data, headers, timeout=self.timeout)
            self.handle_transport_success(url=url)

    def _get_log_message(self, data):
        # decode message so we can show the actual event
        try:
            data = self.decode(data)
        except Exception:
            message = '<failed decoding data>'
        else:
            message = data.pop('message', '<no message value>')
        return message

    def _get_transport(self, parsed_url):
        if self.async_mode and is_master_process():
            # when in the master process, always use SYNC mode. This avoids
            # the danger of being forked into an inconsistent threading state
            self.logger.info('Sending message synchronously while in master '
                             'process. PID: %s', os.getpid())
            return import_string(defaults.SYNC_TRANSPORT_CLASS)(parsed_url)
        if parsed_url not in self._transports:
            self._transports[parsed_url] = self._transport_class(parsed_url)
        return self._transports[parsed_url]

    def _filter_exception_type(self, data):
        exception = data.get('exception')
        if not exception:
            return False

        exc_type = exception.get('type')
        exc_module = exception.get('module')
        if exc_module == 'None':
            exc_module = None

        if exc_type in self.filter_exception_types_dict:
            exc_to_filter_module = self.filter_exception_types_dict[exc_type]
            if not exc_to_filter_module or exc_to_filter_module == exc_module:
                if exc_module:
                    exc_name = '%s.%s' % (exc_module, exc_type)
                else:
                    exc_name = exc_type
                self.logger.info(
                    'Ignored %s exception due to exception type filter',
                    exc_name
                )
                return True
        return False

    def send_remote(self, url, data, headers=None):
        if not self.state.should_try():
            message = self._get_log_message(data)
            self.error_logger.error(message)
            return
        try:
            self._send_remote(url=url, data=data, headers=headers)
        except Exception as e:
            self.handle_transport_fail(exception=e)

    def send(self, secret_token=None, auth_header=None,
             servers=None, **data):
        """
        Serializes the message and passes the payload onto ``send_encoded``.
        """

        if self.is_send_disabled or self._filter_exception_type(data):
            return

        message = self.encode(data)

        return self.send_encoded(message,
                                 secret_token=secret_token,
                                 auth_header=auth_header,
                                 servers=servers)

    def send_encoded(self, message, secret_token, auth_header=None,
                     servers=None, **kwargs):
        """
        Given an already serialized message, signs the message and passes the
        payload off to ``send_remote`` for each server specified in the servers
        configuration.
        """
        servers = servers or self.servers
        if not servers:
            warnings.warn('elasticapm client has no remote servers configured')
            return

        if not auth_header:
            if not secret_token:
                secret_token = self.secret_token

            auth_header = "Bearer %s" % (secret_token)

        for url in servers:
            headers = {
                'Authorization': auth_header,
                'Content-Type': 'application/json',
                'Content-Encoding': 'deflate',
                'User-Agent': 'elasticapm-python/%s' % elasticapm.VERSION,
            }

            self.send_remote(url=url, data=message, headers=headers)

    def encode(self, data):
        """
        Serializes ``data`` into a raw string.
        """
        return zlib.compress(json.dumps(data).encode('utf8'))

    def decode(self, data):
        """
        Unserializes a string, ``data``.
        """
        return json.loads(zlib.decompress(data).decode('utf8'))

    def capture_message(self, message, **kwargs):
        """
        Creates an event from ``message``.

        >>> client.capture_message('My event just happened!')
        """
        return self.capture('Message', message=message, **kwargs)

    @deprecated(alternative="capture_message()")
    def captureMessage(self, message, **kwargs):
        """
        Deprecated
        :param message:
        :type message:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        self.capture_message(message, **kwargs)

    def capture_exception(self, exc_info=None, **kwargs):
        """
        Creates an event from an exception.

        >>> try:
        >>>     exc_info = sys.exc_info()
        >>>     client.capture_exception(exc_info)
        >>> finally:
        >>>     del exc_info

        If exc_info is not provided, or is set to True, then this method will
        perform the ``exc_info = sys.exc_info()`` and the requisite clean-up
        for you.
        """
        return self.capture('Exception', exc_info=exc_info, **kwargs)

    @deprecated(alternative="capture_exception()")
    def captureException(self, exc_info=None, **kwargs):
        """
        Deprecated
        """
        self.capture_exception(exc_info, **kwargs)

    def capture_query(self, query, params=(), engine=None, **kwargs):
        """
        Creates an event for a SQL query.

        >>> client.capture_query('SELECT * FROM foo')
        """
        return self.capture('Query', query=query, params=params, engine=engine,
                            **kwargs)

    @deprecated(alternative="capture_query()")
    def captureQuery(self, *args, **kwargs):
        """
        Deprecated
        """
        self.capture_query(*args, **kwargs)

    def begin_transaction(self, transaction_type):
        """Register the start of a transaction on the client

        'kind' should follow the convention of '<transaction-kind>.<provider>'
        e.g. 'web.django', 'task.celery'.
        """
        self.instrumentation_store.begin_transaction(transaction_type)

    def end_transaction(self, name, status_code=None):
        self.instrumentation_store.end_transaction(status_code, name)
        if self.instrumentation_store.should_collect():
            self._traces_collect()

    def set_transaction_name(self, name):
        transaction = get_transaction()
        if not transaction:
            return
        transaction.name = name

    def set_transaction_extra_data(self, data, _key=None):
        transaction = get_transaction()
        if not transaction:
            return
        if not _key:
            _key = 'extra'
        transaction.extra[_key] = data

    def close(self):
        self._traces_collect()
        for url, transport in self._transports.items():
            transport.close()

    def handle_transport_success(self, **kwargs):
        """
        Success handler called by the transport
        """
        if kwargs.get('url'):
            self.logger.info('Logged error at ' + kwargs['url'])
        self.state.set_success()

    def handle_transport_fail(self, exception=None, **kwargs):
        """
        Failure handler called by the transport
        """
        if isinstance(exception, TransportException):
            message = self._get_log_message(exception.data)
            self.error_logger.error(exception.args[0])
        else:
            # stdlib exception
            message = str(exception)
        self.error_logger.error(
            'Failed to submit message: %r',
            message,
            exc_info=getattr(exception, 'print_trace', True)
        )
        self.state.set_fail()

    def _traces_collect(self):
        transactions = self.instrumentation_store.get_all()
        if not transactions:
            return

        data = self.build_msg({
            'transactions': transactions,
        })
        api_path = defaults.TRANSACTIONS_API_PATH

        self.send(servers=[server + api_path for server in self.servers], **data)

    def get_app_info(self):
        language_version = platform.python_version()
        if hasattr(sys, 'pypy_version_info'):
            runtime_version = '.'.join(map(str, sys.pypy_version_info[:3]))
        else:
            runtime_version = language_version
        return {
            'name': self.app_name,
            'version': self.app_version,
            'agent': {
                'name': 'elasticapm-python',
                'version': elasticapm.VERSION,
            },
            'argv': sys.argv,
            'framework': {
                'name': getattr(self, '_framework', None),
                'version': getattr(self, '_framework_version', None),
            },
            'git_ref': self.git_ref,
            'language': {
                'name': 'python',
                'version': platform.python_version(),
            },
            'pid': os.getpid(),
            'process_title': None,
            'runtime': {
                'name': platform.python_implementation(),
                'version': runtime_version,
            }
        }

    def get_system_info(self):
        return {
            'hostname': socket.gethostname(),
            'architecture': platform.machine(),
            'platform': platform.system().lower(),
        }

class DummyClient(Client):
    """Sends messages into an empty void"""
    def send(self, **kwargs):
        return None



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace
from elasticapm.utils.compat import urlparse


class BotocoreInstrumentation(AbstractInstrumentedModule):
    name = 'botocore'

    instrument_list = [
        ('botocore.client', 'BaseClient._make_api_call'),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        if 'operation_name' in kwargs:
            operation_name = kwargs['operation_name']
        else:
            operation_name = args[0]

        target_endpoint = instance._endpoint.host
        parsed_url = urlparse.urlparse(target_endpoint)
        service, region, _ = parsed_url.hostname.split('.', 2)

        signature = '{}:{}'.format(service, operation_name)
        extra_data = {
            'service': service,
            'region': region,
            'operation': operation_name,
        }

        with trace(signature, 'ext.http.aws', extra_data, leaf=True):
            return wrapped(*args, **kwargs)



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace


class PyLibMcInstrumentation(AbstractInstrumentedModule):
    name = 'pylibmc'

    instrument_list = [
        ("pylibmc", "Client.get"),
        ("pylibmc", "Client.get_multi"),
        ("pylibmc", "Client.set"),
        ("pylibmc", "Client.set_multi"),
        ("pylibmc", "Client.add"),
        ("pylibmc", "Client.replace"),
        ("pylibmc", "Client.append"),
        ("pylibmc", "Client.prepend"),
        ("pylibmc", "Client.incr"),
        ("pylibmc", "Client.decr"),
        ("pylibmc", "Client.gets"),
        ("pylibmc", "Client.cas"),
        ("pylibmc", "Client.delete"),
        ("pylibmc", "Client.delete_multi"),
        ("pylibmc", "Client.touch"),
        ("pylibmc", "Client.get_stats"),
        ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        wrapped_name = self.get_wrapped_name(wrapped, instance, method)
        with trace(wrapped_name, "cache.memcached"):
            return wrapped(*args, **kwargs)



def register_elasticapm(client, worker):
    """Given an ElasticAPM client and an RQ worker, registers exception handlers
    with the worker so exceptions are logged to the apm server.

    E.g.:

    from elasticapm.contrib.django.models import client
    from elasticapm.contrib.rq import register_elasticapm
    
    worker = Worker(map(Queue, listen))
    register_elasticapm(client, worker)
    worker.work()

    """
    def send_to_server(job, *exc_info):
        client.capture_exception(
            exc_info=exc_info,
            extra={
                'job_id': job.id,
                'func': job.func_name,
                'args': job.args,
                'kwargs': job.kwargs,
                'description': job.description,
            }
        )

    worker.push_exc_handler(send_to_server)



"""
elasticapm.utils
~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011-2017 Elasticsearch

Large portions are
:copyright: (c) 2010 by the Sentry Team, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""
import os

from elasticapm.utils import six

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse



default_ports = {
    "https": 433,
    "http": 80,
    "postgresql": 5432
}


def varmap(func, var, context=None, name=None):
    """
    Executes ``func(key_name, value)`` on all values
    recurisively discovering dict and list scoped
    values.
    """
    if context is None:
        context = {}
    objid = id(var)
    if objid in context:
        return func(name, '<...>')
    context[objid] = 1
    if isinstance(var, dict):
        ret = dict((k, varmap(func, v, context, k)) for k, v in six.iteritems(var))
    elif isinstance(var, (list, tuple)):
        ret = [varmap(func, f, context, name) for f in var]
    else:
        ret = func(name, var)
    del context[objid]
    return ret


def disabled_due_to_debug(config, debug):
    """
    Compares module and app configs to determine whether to log to ElasticAPM
    :param config: Dictionary containing module config
    :param debug: Boolean denoting app DEBUG state
    :return: Boolean True if logging is disabled
    """
    return debug and not config.get('DEBUG', False)


def get_name_from_func(func):
    # If no view was set we ignore the request
    module = func.__module__

    if hasattr(func, '__name__'):
        view_name = func.__name__
    else:  # Fall back if there's no __name__
        view_name = func.__class__.__name__

    return '{0}.{1}'.format(module, view_name)


def build_name_with_http_method_prefix(name, request):
    if name:
        return request.method + " " + name
    else:
        return name  # 404


def is_master_process():
    # currently only recognizes uwsgi master process
    try:
        import uwsgi
        return os.getpid() == uwsgi.masterpid()
    except ImportError:
        return False


def get_url_dict(url):
    scheme, netloc, path, params, query, fragment = urlparse.urlparse(url)
    if ':' in netloc:
        hostname, port = netloc.split(':')
    else:
        hostname, port = (netloc, None)
    return {
        'raw': url,
        'protocol': scheme + ':',
        'hostname': hostname,
        'port': port,
        'pathname': path,
        'search': query,
        'hash': fragment,
    }



# -*- coding: utf-8 -*-
import atexit
import functools

try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse

try:
    from urllib2 import HTTPError
except ImportError:
    from urllib.error import HTTPError


def noop_decorator(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapped


def atexit_register(func):
    """
    Uses either uwsgi's atexit mechanism, or atexit from the stdlib.

    When running under uwsgi, using their atexit handler is more reliable,
    especially when using gevent
    :param func: the function to call at exit
    """
    try:
        import uwsgi
        orig = getattr(uwsgi, 'atexit', None)

        def uwsgi_atexit():
            if callable(orig):
                orig()
            func()

        uwsgi.atexit = uwsgi_atexit
    except ImportError:
        atexit.register(func)



from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import trace


class ZLibInstrumentation(AbstractInstrumentedModule):
    name = 'zlib'
    instrument_list = [
        ('zlib', 'compress'),
        ('zlib', 'decompress'),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        wrapped_name = module + "." + method
        with trace(wrapped_name, "compression.zlib"):
            return wrapped(*args, **kwargs)
